# Copyright 2018 Fortinet, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <https://www.gnu.org/licenses/>.

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import os
import json
from ansible_collections.fortinet.fortios.plugins.module_utils.fortimanager.fortimanager import FortiManagerHandler
import pytest

try:
    from ansible_collections.community.fortios.plugins.modules import fmgr_device_provision_template
except ImportError:
    pytest.skip("Could not load required modules for testing", allow_module_level=True)


def load_fixtures():
    fixture_path = os.path.join(os.path.dirname(__file__), 'fixtures') + "/{filename}.json".format(
        filename=os.path.splitext(os.path.basename(__file__))[0])
    try:
        with open(fixture_path, "r") as fixture_file:
            fixture_data = json.load(fixture_file)
    except IOError:
        return []
    return [fixture_data]


@pytest.fixture(autouse=True)
def module_mock(mocker):
    connection_class_mock = mocker.patch('ansible.module_utils.basic.AnsibleModule')
    return connection_class_mock


@pytest.fixture(autouse=True)
def connection_mock(mocker):
    connection_class_mock = mocker.patch('ansible_collections.community.fortios.plugins.modules.fmgr_device_provision_template.Connection')
    return connection_class_mock


@pytest.fixture(scope="function", params=load_fixtures())
def fixture_data(request):
    func_name = request.function.__name__.replace("test_", "")
    return request.param.get(func_name, None)


fmg_instance = FortiManagerHandler(connection_mock, module_mock)


def test_get_devprof(fixture_data, mocker):
    mocker.patch('ansible_collections.fortinet.fortios.plugins.module_utils.fortimanager.fortimanager.FortiManagerHandler.process_request',
                 side_effect=fixture_data)
    #  Fixture sets used:###########################

    ##################################################
    # snmpv3_security_level: None
    # snmp_v2c_query_status: None
    # ntp_type: None
    # dns_suffix: None
    # snmpv3_queries: None
    # snmpv3_trap_status: None
    # snmp_v2c_name: None
    # syslog_facility: syslog
    # snmp_v2c_status: None
    # smtp_validate_cert: None
    # snmpv3_name: None
    # snmp_v2c_trap_src_ipv4: None
    # syslog_port: None
    # ntp_server: None
    # admin_https_port: None
    # ntp_status: None
    # syslog_server: None
    # dns_primary_ipv4: None
    # admin_timeout: None
    # snmpv3_auth_proto: None
    # smtp_port: None
    # snmpv3_priv_pwd: None
    # smtp_server: None
    # syslog_enc_algorithm: disable
    # dns_secondary_ipv4: None
    # smtp_username: None
    # snmpv3_auth_pwd: None
    # syslog_certificate: None
    # admin_fortiguard_target: None
    # snmpv3_query_port: None
    # smtp_password: None
    # adom: ansible
    # snmpv3_notify_hosts: None
    # syslog_mode: udp
    # snmp_v2c_query_port: None
    # snmp_v2c_trap_status: None
    # snmp_status: enable
    # syslog_status: None
    # admin_fortianalyzer_target: None
    # ntp_auth: None
    # snmp_v2c_id: None
    # admin_http_port: None
    # ntp_v3: None
    # snmp_v2c_query_hosts_ipv4: None
    # ntp_sync_interval: None
    # snmpv3_source_ip: None
    # snmpv3_trap_rport: None
    # admin_gui_theme: None
    # syslog_filter: None
    # ntp_auth_pwd: None
    # provisioning_template: ansibleTest
    # smtp_replyto: None
    # provision_targets: None
    # snmp_v2c_trap_port: None
    # snmpv3_priv_proto: None
    # admin_enable_fortiguard: None
    # admin_switch_controller: None
    # admin_language: None
    # smtp_conn_sec: None
    # mode: get
    # smtp_source_ipv4: None
    # snmpv3_status: None
    # delete_provisioning_template: None
    # snmp_v2c_trap_hosts_ipv4: None
    # admin_https_redirect: None
    ##################################################
    ##################################################
    # snmpv3_security_level: None
    # snmp_v2c_query_status: None
    # ntp_type: None
    # dns_suffix: None
    # snmpv3_queries: None
    # snmpv3_trap_status: None
    # snmp_v2c_name: None
    # syslog_facility: kernel
    # snmp_v2c_status: None
    # smtp_validate_cert: None
    # snmpv3_name: None
    # snmp_v2c_trap_src_ipv4: None
    # syslog_port: 514
    # ntp_server: None
    # admin_https_port: None
    # ntp_status: None
    # syslog_server: 10.7.220.59
    # dns_primary_ipv4: None
    # admin_timeout: None
    # snmpv3_auth_proto: None
    # smtp_port: None
    # snmpv3_priv_pwd: None
    # smtp_server: None
    # syslog_enc_algorithm: disable
    # dns_secondary_ipv4: None
    # smtp_replyto: None
    # smtp_username: None
    # snmpv3_auth_pwd: None
    # syslog_certificate: None
    # admin_fortiguard_target: None
    # snmpv3_query_port: None
    # smtp_password: None
    # adom: ansible
    # snmpv3_notify_hosts: None
    # syslog_mode: udp
    # snmp_v2c_query_port: None
    # snmp_v2c_trap_status: None
    # snmp_status: None
    # syslog_status: enable
    # admin_fortianalyzer_target: None
    # ntp_auth: None
    # snmp_v2c_id: None
    # admin_http_port: None
    # snmp_v2c_query_hosts_ipv4: None
    # ntp_sync_interval: None
    # snmpv3_source_ip: None
    # snmpv3_trap_rport: None
    # admin_gui_theme: None
    # syslog_filter: critical
    # ntp_auth_pwd: None
    # provisioning_template: ansibleTest
    # snmp_v2c_trap_hosts_ipv4: None
    # provision_targets: None
    # snmp_v2c_trap_port: None
    # snmpv3_priv_proto: None
    # admin_enable_fortiguard: None
    # admin_switch_controller: None
    # admin_language: None
    # smtp_conn_sec: None
    # mode: get
    # smtp_source_ipv4: None
    # snmpv3_status: None
    # delete_provisioning_template: None
    # ntp_v3: None
    # admin_https_redirect: None
    ##################################################
    ##################################################
    # snmpv3_security_level: None
    # snmp_v2c_query_status: enable
    # ntp_type: None
    # dns_suffix: None
    # snmpv3_queries: None
    # snmpv3_trap_status: None
    # snmp_v2c_name: ansibleV2c
    # syslog_facility: syslog
    # snmp_v2c_status: enable
    # smtp_validate_cert: None
    # snmpv3_name: None
    # snmp_v2c_trap_src_ipv4: 10.7.220.41
    # syslog_port: None
    # ntp_server: None
    # admin_https_port: None
    # ntp_status: None
    # syslog_server: None
    # dns_primary_ipv4: None
    # admin_timeout: None
    # snmpv3_auth_proto: None
    # smtp_port: None
    # snmpv3_priv_pwd: None
    # smtp_server: None
    # syslog_enc_algorithm: disable
    # dns_secondary_ipv4: None
    # smtp_username: None
    # snmpv3_auth_pwd: None
    # syslog_certificate: None
    # admin_fortiguard_target: None
    # snmpv3_query_port: None
    # smtp_password: None
    # adom: ansible
    # snmpv3_notify_hosts: None
    # syslog_mode: udp
    # snmp_v2c_query_port: 162
    # snmp_v2c_trap_status: enable
    # snmp_status: enable
    # syslog_status: None
    # admin_fortianalyzer_target: None
    # ntp_auth: None
    # snmp_v2c_id: 1
    # admin_http_port: None
    # ntp_v3: None
    # snmp_v2c_query_hosts_ipv4: 10.7.220.59 255.255.255.255, 10.7.220.0 255.255.255.0
    # ntp_sync_interval: None
    # snmpv3_source_ip: None
    # snmpv3_trap_rport: None
    # admin_gui_theme: None
    # syslog_filter: None
    # ntp_auth_pwd: None
    # provisioning_template: ansibleTest
    # smtp_replyto: None
    # provision_targets: None
    # snmp_v2c_trap_port: 161
    # snmpv3_priv_proto: None
    # admin_enable_fortiguard: None
    # admin_switch_controller: None
    # admin_language: None
    # smtp_conn_sec: None
    # mode: get
    # smtp_source_ipv4: None
    # snmpv3_status: None
    # delete_provisioning_template: None
    # snmp_v2c_trap_hosts_ipv4: 10.7.220.59 255.255.255.255, 10.7.220.60 255.255.255.255
    # admin_https_redirect: None
    ##################################################
    ##################################################
    # snmpv3_security_level: auth-priv
    # snmp_v2c_query_status: None
    # provision_targets: None
    # ntp_type: None
    # dns_suffix: None
    # snmpv3_queries: enable
    # snmpv3_trap_status: enable
    # snmp_v2c_name: None
    # syslog_facility: syslog
    # snmp_v2c_status: None
    # smtp_validate_cert: None
    # snmpv3_name: ansibleSNMPv3
    # snmp_v2c_id: None
    # syslog_port: None
    # ntp_server: None
    # admin_https_port: None
    # ntp_status: None
    # syslog_server: None
    # admin_switch_controller: None
    # admin_timeout: None
    # snmpv3_auth_proto: sha
    # smtp_port: None
    # snmpv3_priv_pwd: fortinet
    # smtp_server: None
    # syslog_enc_algorithm: disable
    # snmp_v2c_query_hosts_ipv4: None
    # smtp_username: None
    # mode: get
    # syslog_certificate: None
    # admin_fortiguard_target: None
    # snmpv3_query_port: 161
    # smtp_password: None
    # adom: ansible
    # snmpv3_notify_hosts: 10.7.220.59,10.7.220.60
    # syslog_mode: udp
    # snmp_v2c_query_port: None
    # snmp_v2c_trap_status: None
    # snmp_status: enable
    # syslog_status: None
    # snmp_v2c_trap_hosts_ipv4: None
    # admin_fortianalyzer_target: None
    # snmp_v2c_trap_src_ipv4: None
    # admin_http_port: None
    # dns_secondary_ipv4: None
    # syslog_filter: None
    # snmpv3_source_ip: 0.0.0.0
    # snmpv3_trap_rport: 162
    # admin_gui_theme: None
    # ntp_sync_interval: None
    # ntp_auth_pwd: None
    # provisioning_template: ansibleTest
    # ntp_v3: None
    # ntp_auth: None
    # snmp_v2c_trap_port: None
    # snmpv3_priv_proto: aes256
    # admin_enable_fortiguard: None
    # dns_primary_ipv4: None
    # admin_language: None
    # smtp_conn_sec: None
    # snmpv3_auth_pwd: fortinet
    # smtp_source_ipv4: None
    # snmpv3_status: enable
    # delete_provisioning_template: None
    # smtp_replyto: None
    # admin_https_redirect: None
    ##################################################
    ##################################################
    # snmpv3_security_level: None
    # snmp_v2c_query_status: None
    # ntp_type: fortiguard
    # dns_suffix: None
    # snmpv3_queries: None
    # snmpv3_trap_status: None
    # snmp_v2c_name: None
    # syslog_facility: syslog
    # snmp_v2c_status: None
    # smtp_validate_cert: None
    # snmpv3_name: None
    # snmp_v2c_trap_src_ipv4: None
    # syslog_port: None
    # ntp_server: None
    # admin_https_port: None
    # ntp_status: enable
    # syslog_server: None
    # dns_primary_ipv4: None
    # admin_timeout: None
    # snmpv3_auth_proto: None
    # smtp_port: None
    # snmpv3_priv_pwd: None
    # smtp_server: None
    # syslog_enc_algorithm: disable
    # dns_secondary_ipv4: None
    # smtp_replyto: None
    # smtp_username: None
    # snmpv3_auth_pwd: None
    # syslog_certificate: None
    # admin_fortiguard_target: None
    # snmpv3_query_port: None
    # smtp_password: None
    # adom: ansible
    # snmpv3_notify_hosts: None
    # syslog_mode: udp
    # snmp_v2c_query_port: None
    # snmp_v2c_trap_status: None
    # snmp_status: None
    # syslog_status: None
    # admin_fortianalyzer_target: None
    # ntp_auth: None
    # snmp_v2c_id: None
    # admin_http_port: None
    # snmp_v2c_query_hosts_ipv4: None
    # ntp_sync_interval: 60
    # snmpv3_source_ip: None
    # snmpv3_trap_rport: None
    # admin_gui_theme: None
    # syslog_filter: None
    # ntp_auth_pwd: None
    # provisioning_template: ansibleTest
    # snmp_v2c_trap_hosts_ipv4: None
    # provision_targets: None
    # snmp_v2c_trap_port: None
    # snmpv3_priv_proto: None
    # admin_enable_fortiguard: None
    # admin_switch_controller: None
    # admin_language: None
    # smtp_conn_sec: None
    # mode: get
    # smtp_source_ipv4: None
    # snmpv3_status: None
    # delete_provisioning_template: None
    # ntp_v3: None
    # admin_https_redirect: None
    ##################################################
    ##################################################
    # snmpv3_security_level: None
    # snmp_v2c_query_status: None
    # ntp_type: custom
    # dns_suffix: None
    # snmpv3_queries: None
    # snmpv3_trap_status: None
    # snmp_v2c_name: None
    # syslog_facility: syslog
    # snmp_v2c_status: None
    # smtp_validate_cert: None
    # snmpv3_name: None
    # snmp_v2c_trap_src_ipv4: None
    # syslog_port: None
    # ntp_server: 10.7.220.32,10.7.220.1
    # admin_https_port: None
    # ntp_status: enable
    # syslog_server: None
    # dns_primary_ipv4: None
    # admin_timeout: None
    # snmpv3_auth_proto: None
    # smtp_port: None
    # snmpv3_priv_pwd: None
    # smtp_server: None
    # syslog_enc_algorithm: disable
    # dns_secondary_ipv4: None
    # smtp_username: None
    # snmpv3_auth_pwd: None
    # syslog_certificate: None
    # admin_fortiguard_target: None
    # snmpv3_query_port: None
    # smtp_password: None
    # adom: ansible
    # snmpv3_notify_hosts: None
    # syslog_mode: udp
    # snmp_v2c_query_port: None
    # snmp_v2c_trap_status: None
    # snmp_status: None
    # syslog_status: None
    # admin_fortianalyzer_target: None
    # ntp_auth: enable
    # snmp_v2c_id: None
    # admin_http_port: None
    # ntp_v3: None
    # snmp_v2c_query_hosts_ipv4: None
    # ntp_sync_interval: 60
    # snmpv3_source_ip: None
    # snmpv3_trap_rport: None
    # admin_gui_theme: None
    # syslog_filter: None
    # ntp_auth_pwd: fortinet
    # provisioning_template: ansibleTest
    # smtp_replyto: None
    # provision_targets: None
    # snmp_v2c_trap_port: None
    # snmpv3_priv_proto: None
    # admin_enable_fortiguard: None
    # admin_switch_controller: None
    # admin_language: None
    # smtp_conn_sec: None
    # mode: get
    # smtp_source_ipv4: None
    # snmpv3_status: None
    # delete_provisioning_template: None
    # snmp_v2c_trap_hosts_ipv4: None
    # admin_https_redirect: None
    ##################################################
    ##################################################
    # snmpv3_security_level: None
    # snmp_v2c_query_status: None
    # ntp_type: None
    # dns_suffix: None
    # snmpv3_queries: None
    # snmpv3_trap_status: None
    # snmp_v2c_name: None
    # syslog_facility: syslog
    # snmp_v2c_status: None
    # smtp_validate_cert: None
    # snmpv3_name: None
    # snmp_v2c_trap_src_ipv4: None
    # syslog_port: None
    # ntp_server: None
    # admin_https_port: 4433
    # ntp_status: None
    # syslog_server: None
    # dns_primary_ipv4: None
    # admin_timeout: 60
    # snmpv3_auth_proto: None
    # smtp_port: None
    # snmpv3_priv_pwd: None
    # smtp_server: None
    # syslog_enc_algorithm: disable
    # dns_secondary_ipv4: None
    # smtp_replyto: None
    # smtp_username: None
    # snmpv3_auth_pwd: None
    # syslog_certificate: None
    # admin_fortiguard_target: None
    # snmpv3_query_port: None
    # smtp_password: None
    # adom: ansible
    # snmpv3_notify_hosts: None
    # syslog_mode: udp
    # snmp_v2c_query_port: None
    # snmp_v2c_trap_status: None
    # snmp_status: None
    # syslog_status: None
    # admin_fortianalyzer_target: 10.7.220.38
    # ntp_auth: None
    # snmp_v2c_id: None
    # admin_http_port: 8080
    # snmp_v2c_query_hosts_ipv4: None
    # ntp_sync_interval: None
    # snmpv3_source_ip: None
    # snmpv3_trap_rport: None
    # admin_gui_theme: blue
    # syslog_filter: None
    # ntp_auth_pwd: None
    # provisioning_template: ansibleTest
    # snmp_v2c_trap_hosts_ipv4: None
    # provision_targets: None
    # snmp_v2c_trap_port: None
    # snmpv3_priv_proto: None
    # admin_enable_fortiguard: this-fmg
    # admin_switch_controller: enable
    # admin_language: english
    # smtp_conn_sec: None
    # mode: get
    # smtp_source_ipv4: None
    # snmpv3_status: None
    # delete_provisioning_template: None
    # ntp_v3: None
    # admin_https_redirect: enable
    ##################################################
    ##################################################
    # snmpv3_security_level: None
    # snmp_v2c_query_status: None
    # ntp_type: None
    # dns_suffix: None
    # snmpv3_queries: None
    # snmpv3_trap_status: None
    # snmp_v2c_name: None
    # syslog_facility: syslog
    # snmp_v2c_status: None
    # smtp_validate_cert: disable
    # snmpv3_name: None
    # snmp_v2c_trap_src_ipv4: None
    # syslog_port: None
    # ntp_server: None
    # admin_https_port: None
    # ntp_status: None
    # syslog_server: None
    # dns_primary_ipv4: None
    # admin_timeout: None
    # snmpv3_auth_proto: None
    # smtp_port: 25
    # snmpv3_priv_pwd: None
    # smtp_server: 10.7.220.32
    # syslog_enc_algorithm: disable
    # dns_secondary_ipv4: None
    # smtp_username: ansible
    # snmpv3_auth_pwd: None
    # syslog_certificate: None
    # admin_fortiguard_target: None
    # snmpv3_query_port: None
    # smtp_password: fortinet
    # adom: ansible
    # snmpv3_notify_hosts: None
    # syslog_mode: udp
    # snmp_v2c_query_port: None
    # snmp_v2c_trap_status: None
    # snmp_status: None
    # syslog_status: None
    # admin_fortianalyzer_target: None
    # ntp_auth: None
    # snmp_v2c_id: None
    # admin_http_port: None
    # ntp_v3: None
    # snmp_v2c_query_hosts_ipv4: None
    # ntp_sync_interval: None
    # snmpv3_source_ip: None
    # snmpv3_trap_rport: None
    # admin_gui_theme: None
    # syslog_filter: None
    # ntp_auth_pwd: None
    # provisioning_template: ansibleTest
    # smtp_replyto: ansible@do-not-reply.com
    # provision_targets: None
    # snmp_v2c_trap_port: None
    # snmpv3_priv_proto: None
    # admin_enable_fortiguard: None
    # admin_switch_controller: None
    # admin_language: None
    # smtp_conn_sec: starttls
    # mode: get
    # smtp_source_ipv4: 0.0.0.0
    # snmpv3_status: None
    # delete_provisioning_template: None
    # snmp_v2c_trap_hosts_ipv4: None
    # admin_https_redirect: None
    ##################################################
    ##################################################
    # snmpv3_security_level: None
    # snmp_v2c_query_status: None
    # ntp_type: None
    # dns_suffix: ansible.local
    # snmpv3_queries: None
    # snmpv3_trap_status: None
    # snmp_v2c_name: None
    # syslog_facility: syslog
    # snmp_v2c_status: None
    # smtp_validate_cert: None
    # snmpv3_name: None
    # snmp_v2c_trap_src_ipv4: None
    # syslog_port: None
    # ntp_server: None
    # admin_https_port: None
    # ntp_status: None
    # syslog_server: None
    # dns_primary_ipv4: 8.8.8.8
    # admin_timeout: None
    # snmpv3_auth_proto: None
    # smtp_port: None
    # snmpv3_priv_pwd: None
    # smtp_server: None
    # syslog_enc_algorithm: disable
    # dns_secondary_ipv4: 4.4.4.4
    # smtp_replyto: None
    # smtp_username: None
    # snmpv3_auth_pwd: None
    # syslog_certificate: None
    # admin_fortiguard_target: None
    # snmpv3_query_port: None
    # smtp_password: None
    # adom: ansible
    # snmpv3_notify_hosts: None
    # syslog_mode: udp
    # snmp_v2c_query_port: None
    # snmp_v2c_trap_status: None
    # snmp_status: None
    # syslog_status: None
    # admin_fortianalyzer_target: None
    # ntp_auth: None
    # snmp_v2c_id: None
    # admin_http_port: None
    # snmp_v2c_query_hosts_ipv4: None
    # ntp_sync_interval: None
    # snmpv3_source_ip: None
    # snmpv3_trap_rport: None
    # admin_gui_theme: None
    # syslog_filter: None
    # ntp_auth_pwd: None
    # provisioning_template: ansibleTest
    # snmp_v2c_trap_hosts_ipv4: None
    # provision_targets: None
    # snmp_v2c_trap_port: None
    # snmpv3_priv_proto: None
    # admin_enable_fortiguard: None
    # admin_switch_controller: None
    # admin_language: None
    # smtp_conn_sec: None
    # mode: get
    # smtp_source_ipv4: None
    # snmpv3_status: None
    # delete_provisioning_template: None
    # ntp_v3: None
    # admin_https_redirect: None
    ##################################################
    ##################################################
    # snmpv3_security_level: None
    # snmp_v2c_query_status: None
    # ntp_type: None
    # dns_suffix: None
    # snmpv3_queries: None
    # snmpv3_trap_status: None
    # snmp_v2c_name: None
    # syslog_facility: syslog
    # snmp_v2c_status: None
    # smtp_validate_cert: None
    # snmpv3_name: None
    # snmp_v2c_trap_src_ipv4: None
    # syslog_port: None
    # ntp_server: None
    # admin_https_port: None
    # ntp_status: None
    # syslog_server: None
    # dns_primary_ipv4: None
    # admin_timeout: None
    # snmpv3_auth_proto: None
    # smtp_port: None
    # snmpv3_priv_pwd: None
    # smtp_server: None
    # syslog_enc_algorithm: disable
    # dns_secondary_ipv4: None
    # smtp_username: None
    # snmpv3_auth_pwd: None
    # syslog_certificate: None
    # admin_fortiguard_target: None
    # snmpv3_query_port: None
    # smtp_password: None
    # adom: ansible
    # snmpv3_notify_hosts: None
    # syslog_mode: udp
    # snmp_v2c_query_port: None
    # snmp_v2c_trap_status: None
    # snmp_status: None
    # syslog_status: None
    # admin_fortianalyzer_target: None
    # ntp_auth: None
    # snmp_v2c_id: None
    # admin_http_port: None
    # ntp_v3: None
    # snmp_v2c_query_hosts_ipv4: None
    # ntp_sync_interval: None
    # snmpv3_source_ip: None
    # snmpv3_trap_rport: None
    # admin_gui_theme: None
    # syslog_filter: None
    # ntp_auth_pwd: None
    # provisioning_template: ansibleTest
    # smtp_replyto: None
    # provision_targets: FGT1,FGT2
    # snmp_v2c_trap_port: None
    # snmpv3_priv_proto: None
    # admin_enable_fortiguard: None
    # admin_switch_controller: None
    # admin_language: None
    # smtp_conn_sec: None
    # mode: get
    # smtp_source_ipv4: None
    # snmpv3_status: None
    # delete_provisioning_template: None
    # snmp_v2c_trap_hosts_ipv4: None
    # admin_https_redirect: None
    ##################################################

    # Test using fixture 1 #
    output = fmgr_device_provision_template.get_devprof(fmg_instance, fixture_data[0]['paramgram_used'])
    assert isinstance(output['raw_response'], dict) is True
    # Test using fixture 2 #
    output = fmgr_device_provision_template.get_devprof(fmg_instance, fixture_data[1]['paramgram_used'])
    assert isinstance(output['raw_response'], dict) is True
    # Test using fixture 3 #
    output = fmgr_device_provision_template.get_devprof(fmg_instance, fixture_data[2]['paramgram_used'])
    assert isinstance(output['raw_response'], dict) is True
    # Test using fixture 4 #
    output = fmgr_device_provision_template.get_devprof(fmg_instance, fixture_data[3]['paramgram_used'])
    assert isinstance(output['raw_response'], dict) is True
    # Test using fixture 5 #
    output = fmgr_device_provision_template.get_devprof(fmg_instance, fixture_data[4]['paramgram_used'])
    assert isinstance(output['raw_response'], dict) is True
    # Test using fixture 6 #
    output = fmgr_device_provision_template.get_devprof(fmg_instance, fixture_data[5]['paramgram_used'])
    assert isinstance(output['raw_response'], dict) is True
    # Test using fixture 7 #
    output = fmgr_device_provision_template.get_devprof(fmg_instance, fixture_data[6]['paramgram_used'])
    assert isinstance(output['raw_response'], dict) is True
    # Test using fixture 8 #
    output = fmgr_device_provision_template.get_devprof(fmg_instance, fixture_data[7]['paramgram_used'])
    assert isinstance(output['raw_response'], dict) is True
    # Test using fixture 9 #
    output = fmgr_device_provision_template.get_devprof(fmg_instance, fixture_data[8]['paramgram_used'])
    assert isinstance(output['raw_response'], dict) is True
    # Test using fixture 10 #
    output = fmgr_device_provision_template.get_devprof(fmg_instance, fixture_data[9]['paramgram_used'])
    assert isinstance(output['raw_response'], dict) is True


def test_set_devprof(fixture_data, mocker):
    mocker.patch('ansible_collections.fortinet.fortios.plugins.module_utils.fortimanager.fortimanager.FortiManagerHandler.process_request',
                 side_effect=fixture_data)
    #  Fixture sets used:###########################

    ##################################################
    # snmpv3_security_level: None
    # snmp_v2c_query_status: None
    # smtp_port: None
    # ntp_type: None
    # dns_suffix: None
    # snmpv3_queries: None
    # snmpv3_trap_status: None
    # snmp_v2c_name: None
    # syslog_facility: syslog
    # snmp_v2c_status: None
    # smtp_validate_cert: None
    # snmpv3_name: None
    # snmp_v2c_trap_src_ipv4: None
    # syslog_port: None
    # ntp_server: None
    # admin_https_port: None
    # ntp_status: None
    # syslog_server: None
    # dns_primary_ipv4: None
    # admin_timeout: None
    # snmpv3_auth_proto: None
    # ntp_auth: None
    # snmpv3_priv_pwd: None
    # smtp_server: None
    # syslog_enc_algorithm: disable
    # dns_secondary_ipv4: None
    # smtp_replyto: None
    # smtp_username: None
    # snmpv3_auth_pwd: None
    # syslog_certificate: None
    # admin_fortiguard_target: None
    # snmpv3_query_port: None
    # smtp_password: None
    # adom: ansible
    # snmpv3_notify_hosts: None
    # syslog_mode: udp
    # snmp_v2c_query_port: None
    # snmp_v2c_trap_status: None
    # snmp_status: None
    # syslog_status: None
    # admin_fortianalyzer_target: None
    # snmp_v2c_id: None
    # admin_http_port: None
    # snmp_v2c_query_hosts_ipv4: None
    # ntp_sync_interval: None
    # snmpv3_source_ip: None
    # snmpv3_trap_rport: None
    # admin_gui_theme: None
    # syslog_filter: None
    # ntp_auth_pwd: None
    # provisioning_template: None
    # snmp_v2c_trap_hosts_ipv4: None
    # provision_targets: None
    # snmp_v2c_trap_port: None
    # snmpv3_priv_proto: None
    # admin_enable_fortiguard: None
    # admin_switch_controller: None
    # admin_language: None
    # smtp_conn_sec: None
    # mode: delete
    # smtp_source_ipv4: None
    # snmpv3_status: None
    # delete_provisioning_template: ansibleTest
    # ntp_v3: None
    # admin_https_redirect: None
    ##################################################

    # Test using fixture 1 #
    output = fmgr_device_provision_template.set_devprof(fmg_instance, fixture_data[0]['paramgram_used'])
    assert output['raw_response']['status']['code'] == 0


def test_set_devprof_scope(fixture_data, mocker):
    mocker.patch('ansible_collections.fortinet.fortios.plugins.module_utils.fortimanager.fortimanager.FortiManagerHandler.process_request',
                 side_effect=fixture_data)
    #  Fixture sets used:###########################

    ##################################################
    # snmpv3_security_level: None
    # snmp_v2c_query_status: None
    # provision_targets: FGT1,FGT2
    # ntp_type: None
    # dns_suffix: None
    # snmpv3_queries: None
    # snmpv3_trap_status: None
    # snmp_v2c_name: None
    # syslog_facility: syslog
    # snmp_v2c_status: None
    # smtp_validate_cert: None
    # snmpv3_name: None
    # snmp_v2c_id: None
    # syslog_port: None
    # ntp_server: None
    # admin_https_port: None
    # ntp_status: None
    # syslog_server: None
    # admin_switch_controller: None
    # admin_timeout: None
    # snmpv3_auth_proto: None
    # smtp_port: None
    # snmpv3_priv_pwd: None
    # smtp_server: None
    # syslog_enc_algorithm: disable
    # snmp_v2c_query_hosts_ipv4: None
    # smtp_username: None
    # mode: set
    # syslog_certificate: None
    # admin_fortiguard_target: None
    # snmpv3_query_port: None
    # smtp_password: None
    # adom: ansible
    # snmpv3_notify_hosts: None
    # syslog_mode: udp
    # snmp_v2c_query_port: None
    # snmp_v2c_trap_status: None
    # snmp_status: None
    # syslog_status: None
    # admin_https_redirect: None
    # admin_fortianalyzer_target: None
    # snmp_v2c_trap_src_ipv4: None
    # admin_http_port: None
    # dns_secondary_ipv4: None
    # syslog_filter: None
    # snmpv3_source_ip: None
    # snmpv3_trap_rport: None
    # admin_gui_theme: None
    # ntp_sync_interval: None
    # ntp_auth_pwd: None
    # provisioning_template: ansibleTest
    # smtp_replyto: None
    # ntp_auth: None
    # snmp_v2c_trap_port: None
    # snmpv3_priv_proto: None
    # admin_enable_fortiguard: None
    # dns_primary_ipv4: None
    # admin_language: None
    # smtp_conn_sec: None
    # snmpv3_auth_pwd: None
    # smtp_source_ipv4: None
    # snmpv3_status: None
    # delete_provisioning_template: None
    # snmp_v2c_trap_hosts_ipv4: None
    # ntp_v3: None
    ##################################################

    # Test using fixture 1 #
    output = fmgr_device_provision_template.set_devprof_scope(fmg_instance, fixture_data[0]['paramgram_used'])
    assert output['raw_response']['status']['code'] == 0


def test_set_devprof_snmp(fixture_data, mocker):
    mocker.patch('ansible_collections.fortinet.fortios.plugins.module_utils.fortimanager.fortimanager.FortiManagerHandler.process_request',
                 side_effect=fixture_data)
    #  Fixture sets used:###########################

    ##################################################
    # snmpv3_security_level: None
    # snmp_v2c_query_status: None
    # provision_targets: None
    # ntp_type: None
    # dns_suffix: None
    # snmpv3_queries: None
    # snmpv3_trap_status: None
    # snmp_v2c_name: None
    # syslog_facility: syslog
    # snmp_v2c_status: None
    # smtp_validate_cert: None
    # snmpv3_name: None
    # snmp_v2c_id: None
    # syslog_port: None
    # ntp_server: None
    # admin_https_port: None
    # ntp_status: None
    # syslog_server: None
    # admin_switch_controller: None
    # admin_timeout: None
    # snmpv3_auth_proto: None
    # smtp_port: None
    # snmpv3_priv_pwd: None
    # smtp_server: None
    # syslog_enc_algorithm: disable
    # snmp_v2c_query_hosts_ipv4: None
    # smtp_username: None
    # mode: set
    # syslog_certificate: None
    # admin_fortiguard_target: None
    # snmpv3_query_port: None
    # smtp_password: None
    # adom: ansible
    # snmpv3_notify_hosts: None
    # syslog_mode: udp
    # snmp_v2c_query_port: None
    # snmp_v2c_trap_status: None
    # snmp_status: enable
    # syslog_status: None
    # admin_https_redirect: None
    # admin_fortianalyzer_target: None
    # snmp_v2c_trap_src_ipv4: None
    # admin_http_port: None
    # dns_secondary_ipv4: None
    # syslog_filter: None
    # snmpv3_source_ip: None
    # snmpv3_trap_rport: None
    # admin_gui_theme: None
    # ntp_sync_interval: None
    # ntp_auth_pwd: None
    # provisioning_template: ansibleTest
    # smtp_replyto: None
    # ntp_auth: None
    # snmp_v2c_trap_port: None
    # snmpv3_priv_proto: None
    # admin_enable_fortiguard: None
    # dns_primary_ipv4: None
    # admin_language: None
    # smtp_conn_sec: None
    # snmpv3_auth_pwd: None
    # smtp_source_ipv4: None
    # snmpv3_status: None
    # delete_provisioning_template: None
    # snmp_v2c_trap_hosts_ipv4: None
    # ntp_v3: None
    ##################################################
    ##################################################
    # snmpv3_security_level: None
    # snmp_v2c_query_status: enable
    # provision_targets: None
    # ntp_type: None
    # dns_suffix: None
    # snmpv3_queries: None
    # snmpv3_trap_status: None
    # snmp_v2c_name: ansibleV2c
    # syslog_facility: syslog
    # snmp_v2c_status: enable
    # smtp_validate_cert: None
    # snmpv3_name: None
    # snmp_v2c_id: 1
    # syslog_port: None
    # ntp_server: None
    # admin_https_port: None
    # ntp_status: None
    # syslog_server: None
    # admin_switch_controller: None
    # admin_timeout: None
    # snmpv3_auth_proto: None
    # smtp_port: None
    # snmpv3_priv_pwd: None
    # smtp_server: None
    # syslog_enc_algorithm: disable
    # snmp_v2c_query_hosts_ipv4: 10.7.220.59 255.255.255.255, 10.7.220.0 255.255.255.0
    # smtp_username: None
    # mode: set
    # syslog_certificate: None
    # admin_fortiguard_target: None
    # snmpv3_query_port: None
    # smtp_password: None
    # adom: ansible
    # snmpv3_notify_hosts: None
    # syslog_mode: udp
    # snmp_v2c_query_port: 162
    # snmp_v2c_trap_status: enable
    # snmp_status: enable
    # syslog_status: None
    # admin_https_redirect: None
    # admin_fortianalyzer_target: None
    # snmp_v2c_trap_src_ipv4: 10.7.220.41
    # admin_http_port: None
    # dns_secondary_ipv4: None
    # syslog_filter: None
    # snmpv3_source_ip: None
    # snmpv3_trap_rport: None
    # admin_gui_theme: None
    # ntp_sync_interval: None
    # ntp_auth_pwd: None
    # provisioning_template: ansibleTest
    # smtp_replyto: None
    # ntp_auth: None
    # snmp_v2c_trap_port: 161
    # snmpv3_priv_proto: None
    # admin_enable_fortiguard: None
    # dns_primary_ipv4: None
    # admin_language: None
    # smtp_conn_sec: None
    # snmpv3_auth_pwd: None
    # smtp_source_ipv4: None
    # snmpv3_status: None
    # delete_provisioning_template: None
    # snmp_v2c_trap_hosts_ipv4: 10.7.220.59 255.255.255.255, 10.7.220.60 255.255.255.255
    # ntp_v3: None
    ##################################################
    ##################################################
    # snmpv3_security_level: auth-priv
    # snmp_v2c_query_status: None
    # ntp_type: None
    # dns_suffix: None
    # snmpv3_queries: enable
    # snmpv3_trap_status: enable
    # snmp_v2c_name: None
    # syslog_facility: syslog
    # snmp_v2c_status: None
    # smtp_validate_cert: None
    # snmpv3_name: ansibleSNMPv3
    # snmp_v2c_trap_src_ipv4: None
    # syslog_port: None
    # ntp_server: None
    # admin_https_port: None
    # ntp_status: None
    # syslog_server: None
    # dns_primary_ipv4: None
    # admin_timeout: None
    # snmpv3_auth_proto: sha
    # smtp_port: None
    # snmpv3_priv_pwd: fortinet
    # smtp_server: None
    # syslog_enc_algorithm: disable
    # dns_secondary_ipv4: None
    # smtp_username: None
    # snmpv3_auth_pwd: fortinet
    # syslog_certificate: None
    # admin_fortiguard_target: None
    # snmpv3_query_port: 161
    # smtp_password: None
    # adom: ansible
    # snmpv3_notify_hosts: 10.7.220.59,10.7.220.60
    # syslog_mode: udp
    # snmp_v2c_query_port: None
    # snmp_v2c_trap_status: None
    # snmp_status: enable
    # syslog_status: None
    # admin_fortianalyzer_target: None
    # ntp_auth: None
    # snmp_v2c_id: None
    # admin_http_port: None
    # ntp_v3: None
    # snmp_v2c_query_hosts_ipv4: None
    # ntp_sync_interval: None
    # snmpv3_source_ip: 0.0.0.0
    # snmpv3_trap_rport: 162
    # admin_gui_theme: None
    # syslog_filter: None
    # ntp_auth_pwd: None
    # provisioning_template: ansibleTest
    # smtp_replyto: None
    # provision_targets: None
    # snmp_v2c_trap_port: None
    # snmpv3_priv_proto: aes256
    # admin_enable_fortiguard: None
    # admin_switch_controller: None
    # admin_language: None
    # smtp_conn_sec: None
    # mode: set
    # smtp_source_ipv4: None
    # snmpv3_status: enable
    # delete_provisioning_template: None
    # snmp_v2c_trap_hosts_ipv4: None
    # admin_https_redirect: None
    ##################################################

    # Test using fixture 1 #
    output = fmgr_device_provision_template.set_devprof_snmp(fmg_instance, fixture_data[0]['paramgram_used'])
    assert output['raw_response']['status']['code'] == 0
    # Test using fixture 2 #
    output = fmgr_device_provision_template.set_devprof_snmp(fmg_instance, fixture_data[1]['paramgram_used'])
    assert output['raw_response']['status']['code'] == 0
    # Test using fixture 3 #
    output = fmgr_device_provision_template.set_devprof_snmp(fmg_instance, fixture_data[2]['paramgram_used'])
    assert output['raw_response']['status']['code'] == 0


def test_set_devprof_snmp_v2c(fixture_data, mocker):
    mocker.patch('ansible_collections.fortinet.fortios.plugins.module_utils.fortimanager.fortimanager.FortiManagerHandler.process_request',
                 side_effect=fixture_data)
    #  Fixture sets used:###########################

    ##################################################
    # snmpv3_security_level: None
    # snmp_v2c_query_status: enable
    # ntp_type: None
    # dns_suffix: None
    # snmpv3_queries: None
    # snmpv3_trap_status: None
    # snmp_v2c_name: ansibleV2c
    # syslog_facility: syslog
    # snmp_v2c_status: enable
    # smtp_validate_cert: None
    # snmpv3_name: None
    # snmp_v2c_trap_src_ipv4: 10.7.220.41
    # syslog_port: None
    # ntp_server: None
    # admin_https_port: None
    # ntp_status: None
    # syslog_server: None
    # dns_primary_ipv4: None
    # admin_timeout: None
    # snmpv3_auth_proto: None
    # smtp_port: None
    # snmpv3_priv_pwd: None
    # smtp_server: None
    # syslog_enc_algorithm: disable
    # dns_secondary_ipv4: None
    # smtp_replyto: None
    # smtp_username: None
    # snmpv3_auth_pwd: None
    # syslog_certificate: None
    # admin_fortiguard_target: None
    # snmpv3_query_port: None
    # smtp_password: None
    # adom: ansible
    # snmpv3_notify_hosts: None
    # syslog_mode: udp
    # snmp_v2c_query_port: 162
    # snmp_v2c_trap_status: enable
    # snmp_status: enable
    # syslog_status: None
    # admin_fortianalyzer_target: None
    # ntp_auth: None
    # snmp_v2c_id: 1
    # admin_http_port: None
    # snmp_v2c_query_hosts_ipv4: 10.7.220.59 255.255.255.255, 10.7.220.0 255.255.255.0
    # ntp_sync_interval: None
    # snmpv3_source_ip: None
    # snmpv3_trap_rport: None
    # admin_gui_theme: None
    # syslog_filter: None
    # ntp_auth_pwd: None
    # provisioning_template: ansibleTest
    # snmp_v2c_trap_hosts_ipv4: 10.7.220.59 255.255.255.255, 10.7.220.60 255.255.255.255
    # provision_targets: None
    # snmp_v2c_trap_port: 161
    # snmpv3_priv_proto: None
    # admin_enable_fortiguard: None
    # admin_switch_controller: None
    # admin_language: None
    # smtp_conn_sec: None
    # mode: set
    # smtp_source_ipv4: None
    # snmpv3_status: None
    # delete_provisioning_template: None
    # ntp_v3: None
    # admin_https_redirect: None
    ##################################################

    # Test using fixture 1 #
    output = fmgr_device_provision_template.set_devprof_snmp_v2c(fmg_instance, fixture_data[0]['paramgram_used'])
    assert isinstance(output['raw_response'], dict) is True


def test_set_devprof_snmp_v3(fixture_data, mocker):
    mocker.patch('ansible_collections.fortinet.fortios.plugins.module_utils.fortimanager.fortimanager.FortiManagerHandler.process_request',
                 side_effect=fixture_data)
    #  Fixture sets used:###########################

    ##################################################
    # snmpv3_security_level: auth-priv
    # snmp_v2c_query_status: None
    # provision_targets: None
    # ntp_type: None
    # dns_suffix: None
    # snmpv3_queries: enable
    # snmpv3_trap_status: enable
    # snmp_v2c_name: None
    # syslog_facility: syslog
    # snmp_v2c_status: None
    # smtp_validate_cert: None
    # snmpv3_name: ansibleSNMPv3
    # snmp_v2c_id: None
    # syslog_port: None
    # ntp_server: None
    # admin_https_port: None
    # ntp_status: None
    # syslog_server: None
    # admin_switch_controller: None
    # admin_timeout: None
    # snmpv3_auth_proto: sha
    # smtp_port: None
    # snmpv3_priv_pwd: fortinet
    # smtp_server: None
    # syslog_enc_algorithm: disable
    # snmp_v2c_query_hosts_ipv4: None
    # smtp_username: None
    # mode: set
    # syslog_certificate: None
    # admin_fortiguard_target: None
    # snmpv3_query_port: 161
    # smtp_password: None
    # adom: ansible
    # snmpv3_notify_hosts: 10.7.220.59,10.7.220.60
    # syslog_mode: udp
    # snmp_v2c_query_port: None
    # snmp_v2c_trap_status: None
    # snmp_status: enable
    # syslog_status: None
    # admin_https_redirect: None
    # admin_fortianalyzer_target: None
    # snmp_v2c_trap_src_ipv4: None
    # admin_http_port: None
    # dns_secondary_ipv4: None
    # syslog_filter: None
    # snmpv3_source_ip: 0.0.0.0
    # snmpv3_trap_rport: 162
    # admin_gui_theme: None
    # ntp_sync_interval: None
    # ntp_auth_pwd: None
    # provisioning_template: ansibleTest
    # smtp_replyto: None
    # ntp_auth: None
    # snmp_v2c_trap_port: None
    # snmpv3_priv_proto: aes256
    # admin_enable_fortiguard: None
    # dns_primary_ipv4: None
    # admin_language: None
    # smtp_conn_sec: None
    # snmpv3_auth_pwd: fortinet
    # smtp_source_ipv4: None
    # snmpv3_status: enable
    # delete_provisioning_template: None
    # snmp_v2c_trap_hosts_ipv4: None
    # ntp_v3: None
    ##################################################

    # Test using fixture 1 #
    output = fmgr_device_provision_template.set_devprof_snmp_v3(fmg_instance, fixture_data[0]['paramgram_used'])
    assert output['raw_response']['status']['code'] == 0


def test_set_devprof_syslog(fixture_data, mocker):
    mocker.patch('ansible_collections.fortinet.fortios.plugins.module_utils.fortimanager.fortimanager.FortiManagerHandler.process_request',
                 side_effect=fixture_data)
    #  Fixture sets used:###########################

    ##################################################
    # snmpv3_security_level: None
    # snmp_v2c_query_status: None
    # provision_targets: None
    # ntp_type: None
    # dns_suffix: None
    # snmpv3_queries: None
    # snmpv3_trap_status: None
    # snmp_v2c_name: None
    # syslog_facility: kernel
    # snmp_v2c_status: None
    # smtp_validate_cert: None
    # snmpv3_name: None
    # snmp_v2c_id: None
    # syslog_port: 514
    # ntp_server: None
    # admin_https_port: None
    # ntp_status: None
    # syslog_server: 10.7.220.59
    # admin_switch_controller: None
    # admin_timeout: None
    # snmpv3_auth_proto: None
    # smtp_port: None
    # snmpv3_priv_pwd: None
    # smtp_server: None
    # syslog_enc_algorithm: disable
    # snmp_v2c_query_hosts_ipv4: None
    # smtp_username: None
    # mode: set
    # syslog_certificate: None
    # admin_fortiguard_target: None
    # snmpv3_query_port: None
    # smtp_password: None
    # adom: ansible
    # snmpv3_notify_hosts: None
    # syslog_mode: udp
    # snmp_v2c_query_port: None
    # snmp_v2c_trap_status: None
    # snmp_status: None
    # syslog_status: enable
    # snmp_v2c_trap_hosts_ipv4: None
    # admin_fortianalyzer_target: None
    # snmp_v2c_trap_src_ipv4: None
    # admin_http_port: None
    # dns_secondary_ipv4: None
    # syslog_filter: critical
    # snmpv3_source_ip: None
    # snmpv3_trap_rport: None
    # admin_gui_theme: None
    # ntp_sync_interval: None
    # ntp_auth_pwd: None
    # provisioning_template: ansibleTest
    # ntp_v3: None
    # ntp_auth: None
    # snmp_v2c_trap_port: None
    # snmpv3_priv_proto: None
    # admin_enable_fortiguard: None
    # dns_primary_ipv4: None
    # admin_language: None
    # smtp_conn_sec: None
    # snmpv3_auth_pwd: None
    # smtp_source_ipv4: None
    # snmpv3_status: None
    # delete_provisioning_template: None
    # smtp_replyto: None
    # admin_https_redirect: None
    ##################################################

    # Test using fixture 1 #
    output = fmgr_device_provision_template.set_devprof_syslog(fmg_instance, fixture_data[0]['paramgram_used'])
    assert output['raw_response']['status']['code'] == 0


def test_set_devprof_ntp(fixture_data, mocker):
    mocker.patch('ansible_collections.fortinet.fortios.plugins.module_utils.fortimanager.fortimanager.FortiManagerHandler.process_request',
                 side_effect=fixture_data)
    #  Fixture sets used:###########################

    ##################################################
    # snmpv3_security_level: None
    # snmp_v2c_query_status: None
    # provision_targets: None
    # ntp_type: fortiguard
    # dns_suffix: None
    # snmpv3_queries: None
    # snmpv3_trap_status: None
    # snmp_v2c_name: None
    # syslog_facility: syslog
    # snmp_v2c_status: None
    # smtp_validate_cert: None
    # snmpv3_name: None
    # snmp_v2c_id: None
    # syslog_port: None
    # ntp_server: None
    # admin_https_port: None
    # ntp_status: enable
    # syslog_server: None
    # admin_switch_controller: None
    # admin_timeout: None
    # snmpv3_auth_proto: None
    # smtp_port: None
    # snmpv3_priv_pwd: None
    # smtp_server: None
    # syslog_enc_algorithm: disable
    # snmp_v2c_query_hosts_ipv4: None
    # smtp_username: None
    # mode: set
    # syslog_certificate: None
    # admin_fortiguard_target: None
    # snmpv3_query_port: None
    # smtp_password: None
    # adom: ansible
    # snmpv3_notify_hosts: None
    # syslog_mode: udp
    # snmp_v2c_query_port: None
    # snmp_v2c_trap_status: None
    # snmp_status: None
    # syslog_status: None
    # snmp_v2c_trap_hosts_ipv4: None
    # admin_fortianalyzer_target: None
    # snmp_v2c_trap_src_ipv4: None
    # admin_http_port: None
    # dns_secondary_ipv4: None
    # syslog_filter: None
    # snmpv3_source_ip: None
    # snmpv3_trap_rport: None
    # admin_gui_theme: None
    # ntp_sync_interval: 60
    # ntp_auth_pwd: None
    # provisioning_template: ansibleTest
    # ntp_v3: None
    # ntp_auth: None
    # snmp_v2c_trap_port: None
    # snmpv3_priv_proto: None
    # admin_enable_fortiguard: None
    # dns_primary_ipv4: None
    # admin_language: None
    # smtp_conn_sec: None
    # snmpv3_auth_pwd: None
    # smtp_source_ipv4: None
    # snmpv3_status: None
    # delete_provisioning_template: None
    # smtp_replyto: None
    # admin_https_redirect: None
    ##################################################
    ##################################################
    # snmpv3_security_level: None
    # snmp_v2c_query_status: None
    # provision_targets: None
    # ntp_type: custom
    # dns_suffix: None
    # snmpv3_queries: None
    # snmpv3_trap_status: None
    # snmp_v2c_name: None
    # syslog_facility: syslog
    # snmp_v2c_status: None
    # smtp_validate_cert: None
    # snmpv3_name: None
    # snmp_v2c_id: None
    # syslog_port: None
    # ntp_server: 10.7.220.32,10.7.220.1
    # admin_https_port: None
    # ntp_status: enable
    # syslog_server: None
    # admin_switch_controller: None
    # admin_timeout: None
    # snmpv3_auth_proto: None
    # smtp_port: None
    # snmpv3_priv_pwd: None
    # smtp_server: None
    # syslog_enc_algorithm: disable
    # snmp_v2c_query_hosts_ipv4: None
    # smtp_username: None
    # mode: set
    # syslog_certificate: None
    # admin_fortiguard_target: None
    # snmpv3_query_port: None
    # smtp_password: None
    # adom: ansible
    # snmpv3_notify_hosts: None
    # syslog_mode: udp
    # snmp_v2c_query_port: None
    # snmp_v2c_trap_status: None
    # snmp_status: None
    # syslog_status: None
    # admin_https_redirect: None
    # admin_fortianalyzer_target: None
    # snmp_v2c_trap_src_ipv4: None
    # admin_http_port: None
    # dns_secondary_ipv4: None
    # syslog_filter: None
    # snmpv3_source_ip: None
    # snmpv3_trap_rport: None
    # admin_gui_theme: None
    # ntp_sync_interval: 60
    # ntp_auth_pwd: fortinet
    # provisioning_template: ansibleTest
    # smtp_replyto: None
    # ntp_auth: enable
    # snmp_v2c_trap_port: None
    # snmpv3_priv_proto: None
    # admin_enable_fortiguard: None
    # dns_primary_ipv4: None
    # admin_language: None
    # smtp_conn_sec: None
    # snmpv3_auth_pwd: None
    # smtp_source_ipv4: None
    # snmpv3_status: None
    # delete_provisioning_template: None
    # snmp_v2c_trap_hosts_ipv4: None
    # ntp_v3: None
    ##################################################

    # Test using fixture 1 #
    output = fmgr_device_provision_template.set_devprof_ntp(fmg_instance, fixture_data[0]['paramgram_used'])
    assert output['raw_response']['status']['code'] == 0
    # Test using fixture 2 #
    output = fmgr_device_provision_template.set_devprof_ntp(fmg_instance, fixture_data[1]['paramgram_used'])
    assert output['raw_response']['status']['code'] == 0


def test_set_devprof_admin(fixture_data, mocker):
    mocker.patch('ansible_collections.fortinet.fortios.plugins.module_utils.fortimanager.fortimanager.FortiManagerHandler.process_request',
                 side_effect=fixture_data)
    #  Fixture sets used:###########################

    ##################################################
    # snmpv3_security_level: None
    # snmp_v2c_query_status: None
    # provision_targets: None
    # ntp_type: None
    # dns_suffix: None
    # snmpv3_queries: None
    # snmpv3_trap_status: None
    # snmp_v2c_name: None
    # syslog_facility: syslog
    # snmp_v2c_status: None
    # smtp_validate_cert: None
    # snmpv3_name: None
    # snmp_v2c_id: None
    # syslog_port: None
    # ntp_server: None
    # admin_https_port: 4433
    # ntp_status: None
    # syslog_server: None
    # admin_switch_controller: enable
    # admin_timeout: 60
    # snmpv3_auth_proto: None
    # smtp_port: None
    # snmpv3_priv_pwd: None
    # smtp_server: None
    # syslog_enc_algorithm: disable
    # snmp_v2c_query_hosts_ipv4: None
    # smtp_username: None
    # mode: set
    # syslog_certificate: None
    # admin_fortiguard_target: None
    # snmpv3_query_port: None
    # smtp_password: None
    # adom: ansible
    # snmpv3_notify_hosts: None
    # syslog_mode: udp
    # snmp_v2c_query_port: None
    # snmp_v2c_trap_status: None
    # snmp_status: None
    # syslog_status: None
    # snmp_v2c_trap_hosts_ipv4: None
    # admin_fortianalyzer_target: 10.7.220.38
    # snmp_v2c_trap_src_ipv4: None
    # admin_http_port: 8080
    # dns_secondary_ipv4: None
    # syslog_filter: None
    # snmpv3_source_ip: None
    # snmpv3_trap_rport: None
    # admin_gui_theme: blue
    # ntp_sync_interval: None
    # ntp_auth_pwd: None
    # provisioning_template: ansibleTest
    # ntp_v3: None
    # ntp_auth: None
    # snmp_v2c_trap_port: None
    # snmpv3_priv_proto: None
    # admin_enable_fortiguard: this-fmg
    # dns_primary_ipv4: None
    # admin_language: english
    # smtp_conn_sec: None
    # snmpv3_auth_pwd: None
    # smtp_source_ipv4: None
    # snmpv3_status: None
    # delete_provisioning_template: None
    # smtp_replyto: None
    # admin_https_redirect: enable
    ##################################################

    # Test using fixture 1 #
    output = fmgr_device_provision_template.set_devprof_admin(fmg_instance, fixture_data[0]['paramgram_used'])
    assert output['raw_response']['status']['code'] == 0


def test_set_devprof_smtp(fixture_data, mocker):
    mocker.patch('ansible_collections.fortinet.fortios.plugins.module_utils.fortimanager.fortimanager.FortiManagerHandler.process_request',
                 side_effect=fixture_data)
    #  Fixture sets used:###########################

    ##################################################
    # snmpv3_security_level: None
    # snmp_v2c_query_status: None
    # provision_targets: None
    # ntp_type: None
    # dns_suffix: None
    # snmpv3_queries: None
    # snmpv3_trap_status: None
    # snmp_v2c_name: None
    # syslog_facility: syslog
    # snmp_v2c_status: None
    # smtp_validate_cert: disable
    # snmpv3_name: None
    # snmp_v2c_id: None
    # syslog_port: None
    # ntp_server: None
    # admin_https_port: None
    # ntp_status: None
    # syslog_server: None
    # admin_switch_controller: None
    # admin_timeout: None
    # snmpv3_auth_proto: None
    # smtp_port: 25
    # snmpv3_priv_pwd: None
    # smtp_server: 10.7.220.32
    # syslog_enc_algorithm: disable
    # snmp_v2c_query_hosts_ipv4: None
    # smtp_username: ansible
    # mode: set
    # syslog_certificate: None
    # admin_fortiguard_target: None
    # snmpv3_query_port: None
    # smtp_password: fortinet
    # adom: ansible
    # snmpv3_notify_hosts: None
    # syslog_mode: udp
    # snmp_v2c_query_port: None
    # snmp_v2c_trap_status: None
    # snmp_status: None
    # syslog_status: None
    # admin_https_redirect: None
    # admin_fortianalyzer_target: None
    # snmp_v2c_trap_src_ipv4: None
    # admin_http_port: None
    # dns_secondary_ipv4: None
    # syslog_filter: None
    # snmpv3_source_ip: None
    # snmpv3_trap_rport: None
    # admin_gui_theme: None
    # ntp_sync_interval: None
    # ntp_auth_pwd: None
    # provisioning_template: ansibleTest
    # smtp_replyto: ansible@do-not-reply.com
    # ntp_auth: None
    # snmp_v2c_trap_port: None
    # snmpv3_priv_proto: None
    # admin_enable_fortiguard: None
    # dns_primary_ipv4: None
    # admin_language: None
    # smtp_conn_sec: starttls
    # snmpv3_auth_pwd: None
    # smtp_source_ipv4: 0.0.0.0
    # snmpv3_status: None
    # delete_provisioning_template: None
    # snmp_v2c_trap_hosts_ipv4: None
    # ntp_v3: None
    ##################################################

    # Test using fixture 1 #
    output = fmgr_device_provision_template.set_devprof_smtp(fmg_instance, fixture_data[0]['paramgram_used'])
    assert output['raw_response']['status']['code'] == 0


def test_set_devprof_dns(fixture_data, mocker):
    mocker.patch('ansible_collections.fortinet.fortios.plugins.module_utils.fortimanager.fortimanager.FortiManagerHandler.process_request',
                 side_effect=fixture_data)
    #  Fixture sets used:###########################

    ##################################################
    # snmpv3_security_level: None
    # snmp_v2c_query_status: None
    # provision_targets: None
    # ntp_type: None
    # dns_suffix: ansible.local
    # snmpv3_queries: None
    # snmpv3_trap_status: None
    # snmp_v2c_name: None
    # syslog_facility: syslog
    # snmp_v2c_status: None
    # smtp_validate_cert: None
    # snmpv3_name: None
    # snmp_v2c_id: None
    # syslog_port: None
    # ntp_server: None
    # admin_https_port: None
    # ntp_status: None
    # syslog_server: None
    # admin_switch_controller: None
    # admin_timeout: None
    # snmpv3_auth_proto: None
    # smtp_port: None
    # snmpv3_priv_pwd: None
    # smtp_server: None
    # syslog_enc_algorithm: disable
    # snmp_v2c_query_hosts_ipv4: None
    # smtp_username: None
    # mode: set
    # syslog_certificate: None
    # admin_fortiguard_target: None
    # snmpv3_query_port: None
    # smtp_password: None
    # adom: ansible
    # snmpv3_notify_hosts: None
    # syslog_mode: udp
    # snmp_v2c_query_port: None
    # snmp_v2c_trap_status: None
    # snmp_status: None
    # syslog_status: None
    # snmp_v2c_trap_hosts_ipv4: None
    # admin_fortianalyzer_target: None
    # snmp_v2c_trap_src_ipv4: None
    # admin_http_port: None
    # dns_secondary_ipv4: 4.4.4.4
    # syslog_filter: None
    # snmpv3_source_ip: None
    # snmpv3_trap_rport: None
    # admin_gui_theme: None
    # ntp_sync_interval: None
    # ntp_auth_pwd: None
    # provisioning_template: ansibleTest
    # ntp_v3: None
    # ntp_auth: None
    # snmp_v2c_trap_port: None
    # snmpv3_priv_proto: None
    # admin_enable_fortiguard: None
    # dns_primary_ipv4: 8.8.8.8
    # admin_language: None
    # smtp_conn_sec: None
    # snmpv3_auth_pwd: None
    # smtp_source_ipv4: None
    # snmpv3_status: None
    # delete_provisioning_template: None
    # smtp_replyto: None
    # admin_https_redirect: None
    ##################################################

    # Test using fixture 1 #
    output = fmgr_device_provision_template.set_devprof_dns(fmg_instance, fixture_data[0]['paramgram_used'])
    assert output['raw_response']['status']['code'] == 0



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2020, Zainab Alsaffar <Zainab.Alsaffar@mail.rit.edu>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = r'''
---
module: gitlab_group_members
short_description: Manage group members on GitLab Server
description:
    - This module allows to add and remove members to/from a group, or change a member's access level in a group on GitLab.
version_added: '1.2.0'
author: Zainab Alsaffar (@zanssa)
requirements:
    - python-gitlab python module <= 1.15.0
    - administrator rights on the GitLab server
extends_documentation_fragment: community.general.auth_basic
options:
    api_token:
        description:
            - A personal access token to authenticate with the GitLab API.
        required: true
        type: str
    gitlab_group:
        description:
            - The name of the GitLab group the member is added to/removed from.
        required: true
        type: str
    gitlab_user:
        description:
            - The username of the member to add to/remove from the GitLab group.
        required: true
        type: str
    access_level:
        description:
            - The access level for the user.
            - Required if I(state=present), user state is set to present.
        type: str
        choices: ['guest', 'reporter', 'developer', 'maintainer', 'owner']
    state:
        description:
            - State of the member in the group.
            - On C(present), it adds a user to a GitLab group.
            - On C(absent), it removes a user from a GitLab group.
        choices: ['present', 'absent']
        default: 'present'
        type: str
notes:
    - Supports C(check_mode).
'''

EXAMPLES = r'''
- name: Add a user to a GitLab Group
  community.general.gitlab_group_members:
    api_url: 'https://gitlab.example.com'
    api_token: 'Your-Private-Token'
    gitlab_group: groupname
    gitlab_user: username
    access_level: developer
    state: present

- name: Remove a user from a GitLab Group
  community.general.gitlab_group_members:
    api_url: 'https://gitlab.example.com'
    api_token: 'Your-Private-Token'
    gitlab_group: groupname
    gitlab_user: username
    state: absent
'''

RETURN = r''' # '''

from ansible.module_utils.api import basic_auth_argument_spec
from ansible.module_utils.basic import AnsibleModule, missing_required_lib

from ansible_collections.community.general.plugins.module_utils.gitlab import gitlabAuthentication

import traceback

try:
    import gitlab
    HAS_PY_GITLAB = True
except ImportError:
    GITLAB_IMP_ERR = traceback.format_exc()
    HAS_PY_GITLAB = False


class GitLabGroup(object):
    def __init__(self, module, gl):
        self._module = module
        self._gitlab = gl

    # get user id if the user exists
    def get_user_id(self, gitlab_user):
        user_exists = self._gitlab.users.list(username=gitlab_user)
        if user_exists:
            return user_exists[0].id

    # get group id if group exists
    def get_group_id(self, gitlab_group):
        group_exists = self._gitlab.groups.list(search=gitlab_group)
        if group_exists:
            return group_exists[0].id

    # get all members in a group
    def get_members_in_a_group(self, gitlab_group_id):
        group = self._gitlab.groups.get(gitlab_group_id)
        return group.members.list()

    # check if the user is a member of the group
    def is_user_a_member(self, members, gitlab_user_id):
        for member in members:
            if member.id == gitlab_user_id:
                return True
        return False

    # add user to a group
    def add_member_to_group(self, gitlab_user_id, gitlab_group_id, access_level):
        try:
            group = self._gitlab.groups.get(gitlab_group_id)
            add_member = group.members.create(
                {'user_id': gitlab_user_id, 'access_level': access_level})

            if add_member:
                return add_member.username

        except (gitlab.exceptions.GitlabCreateError) as e:
            self._module.fail_json(
                msg="Failed to add member to the Group, Group ID %s: %s" % (gitlab_group_id, e))

    # remove user from a group
    def remove_user_from_group(self, gitlab_user_id, gitlab_group_id):
        try:
            group = self._gitlab.groups.get(gitlab_group_id)
            group.members.delete(gitlab_user_id)

        except (gitlab.exceptions.GitlabDeleteError) as e:
            self._module.fail_json(
                msg="Failed to remove member from GitLab group, ID %s: %s" % (gitlab_group_id, e))

    # get user's access level
    def get_user_access_level(self, members, gitlab_user_id):
        for member in members:
            if member.id == gitlab_user_id:
                return member.access_level

    # update user's access level in a group
    def update_user_access_level(self, members, gitlab_user_id, access_level):
        for member in members:
            if member.id == gitlab_user_id:
                try:
                    member.access_level = access_level
                    member.save()
                except (gitlab.exceptions.GitlabCreateError) as e:
                    self._module.fail_json(
                        msg="Failed to update the access level for the member, %s: %s" % (gitlab_user_id, e))


def main():
    argument_spec = basic_auth_argument_spec()
    argument_spec.update(dict(
        api_token=dict(type='str', required=True, no_log=True),
        gitlab_group=dict(type='str', required=True),
        gitlab_user=dict(type='str', required=True),
        state=dict(type='str', default='present', choices=['present', 'absent']),
        access_level=dict(type='str', required=False, choices=['guest', 'reporter', 'developer', 'maintainer', 'owner'])
    ))

    module = AnsibleModule(
        argument_spec=argument_spec,
        mutually_exclusive=[
            ['api_username', 'api_token'],
            ['api_password', 'api_token'],
        ],
        required_together=[
            ['api_username', 'api_password'],
        ],
        required_one_of=[
            ['api_username', 'api_token'],
        ],
        required_if=[
            ['state', 'present', ['access_level']],
        ],
        supports_check_mode=True,
    )

    if not HAS_PY_GITLAB:
        module.fail_json(msg=missing_required_lib('python-gitlab', url='https://python-gitlab.readthedocs.io/en/stable/'), exception=GITLAB_IMP_ERR)

    gitlab_group = module.params['gitlab_group']
    gitlab_user = module.params['gitlab_user']
    state = module.params['state']
    access_level = module.params['access_level']

    # convert access level string input to int
    if access_level:
        access_level_int = {
            'guest': gitlab.GUEST_ACCESS,
            'reporter': gitlab.REPORTER_ACCESS,
            'developer': gitlab.DEVELOPER_ACCESS,
            'maintainer': gitlab.MAINTAINER_ACCESS,
            'owner': gitlab.OWNER_ACCESS
        }

        access_level = access_level_int[access_level]

    # connect to gitlab server
    gl = gitlabAuthentication(module)

    group = GitLabGroup(module, gl)

    gitlab_user_id = group.get_user_id(gitlab_user)
    gitlab_group_id = group.get_group_id(gitlab_group)

    # group doesn't exist
    if not gitlab_group_id:
        module.fail_json(msg="group '%s' not found." % gitlab_group)

    # user doesn't exist
    if not gitlab_user_id:
        if state == 'absent':
            module.exit_json(changed=False, result="user '%s' not found, and thus also not part of the group" % gitlab_user)
        else:
            module.fail_json(msg="user '%s' not found." % gitlab_user)

    members = group.get_members_in_a_group(gitlab_group_id)
    is_user_a_member = group.is_user_a_member(members, gitlab_user_id)

    # check if the user is a member in the group
    if not is_user_a_member:
        if state == 'present':
            # add user to the group
            if not module.check_mode:
                group.add_member_to_group(gitlab_user_id, gitlab_group_id, access_level)
            module.exit_json(changed=True, result="Successfully added user '%s' to the group." % gitlab_user)
        # state as absent
        else:
            module.exit_json(changed=False, result="User, '%s', is not a member in the group. No change to report" % gitlab_user)
    # in case that a user is a member
    else:
        if state == 'present':
            # compare the access level
            user_access_level = group.get_user_access_level(members, gitlab_user_id)
            if user_access_level == access_level:
                module.exit_json(changed=False, result="User, '%s', is already a member in the group. No change to report" % gitlab_user)
            else:
                # update the access level for the user
                if not module.check_mode:
                    group.update_user_access_level(members, gitlab_user_id, access_level)
                module.exit_json(changed=True, result="Successfully updated the access level for the user, '%s'" % gitlab_user)
        else:
            # remove the user from the group
            if not module.check_mode:
                group.remove_user_from_group(gitlab_user_id, gitlab_group_id)
            module.exit_json(changed=True, result="Successfully removed user, '%s', from the group" % gitlab_user)


if __name__ == '__main__':
    main()



#!/usr/bin/python
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
---
module: fmgr_ha
notes:
    - Full Documentation at U(https://ftnt-ansible-docs.readthedocs.io/en/latest/).
author:
    - Luke Weighall (@lweighall)
    - Andrew Welsh (@Ghilli3)
    - Jim Huber (@p4r4n0y1ng)
short_description: Manages the High-Availability State of FortiManager Clusters and Nodes.
description: Change HA state or settings of FortiManager nodes (Standalone/Master/Slave).

options:
  fmgr_ha_mode:
    description:
      - Sets the role of the FortiManager host for HA.
    required: false
    choices: ["standalone", "master", "slave"]

  fmgr_ha_peer_ipv4:
    description:
      - Sets the IPv4 address of a HA peer.
    required: false

  fmgr_ha_peer_ipv6:
    description:
      - Sets the IPv6 address of a HA peer.
    required: false

  fmgr_ha_peer_sn:
    description:
      - Sets the HA Peer Serial Number.
    required: false

  fmgr_ha_peer_status:
    description:
      - Sets the peer status to enable or disable.
    required: false
    choices: ["enable", "disable"]

  fmgr_ha_cluster_pw:
    description:
      - Sets the password for the HA cluster. Only required once. System remembers between HA mode switches.
    required: false

  fmgr_ha_cluster_id:
    description:
      - Sets the ID number of the HA cluster. Defaults to 1.
    required: false
    default: 1

  fmgr_ha_hb_threshold:
    description:
      - Sets heartbeat lost threshold (1-255).
    required: false
    default: 3

  fmgr_ha_hb_interval:
    description:
      - Sets the heartbeat interval (1-255).
    required: false
    default: 5

  fmgr_ha_file_quota:
    description:
      - Sets the File quota in MB (2048-20480).
    required: false
    default: 4096
'''


EXAMPLES = '''
- name: SET FORTIMANAGER HA NODE TO MASTER
  community.fortios.fmgr_ha:
    fmgr_ha_mode: "master"
    fmgr_ha_cluster_pw: "fortinet"
    fmgr_ha_cluster_id: "1"

- name: SET FORTIMANAGER HA NODE TO SLAVE
  community.fortios.fmgr_ha:
    fmgr_ha_mode: "slave"
    fmgr_ha_cluster_pw: "fortinet"
    fmgr_ha_cluster_id: "1"

- name: SET FORTIMANAGER HA NODE TO STANDALONE
  community.fortios.fmgr_ha:
    fmgr_ha_mode: "standalone"

- name: ADD FORTIMANAGER HA PEER
  community.fortios.fmgr_ha:
    fmgr_ha_peer_ipv4: "192.168.1.254"
    fmgr_ha_peer_sn: "FMG-VM1234567890"
    fmgr_ha_peer_status: "enable"

- name: CREATE CLUSTER ON MASTER
  community.fortios.fmgr_ha:
    fmgr_ha_mode: "master"
    fmgr_ha_cluster_pw: "fortinet"
    fmgr_ha_cluster_id: "1"
    fmgr_ha_hb_threshold: "10"
    fmgr_ha_hb_interval: "15"
    fmgr_ha_file_quota: "2048"
'''
RETURN = """
api_result:
  description: full API response, includes status code and message
  returned: always
  type: str
"""

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.connection import Connection
from ansible_collections.fortinet.fortios.plugins.module_utils.fortimanager.fortimanager import FortiManagerHandler
from ansible_collections.fortinet.fortios.plugins.module_utils.fortimanager.common import FMGBaseException
from ansible_collections.fortinet.fortios.plugins.module_utils.fortimanager.common import FMGRCommon
from ansible_collections.fortinet.fortios.plugins.module_utils.fortimanager.common import FMGRMethods
from ansible_collections.fortinet.fortios.plugins.module_utils.fortimanager.common import DEFAULT_RESULT_OBJ
from ansible_collections.fortinet.fortios.plugins.module_utils.fortimanager.common import FAIL_SOCKET_MSG


def fmgr_set_ha_mode(fmgr, paramgram):
    """
    :param fmgr: The fmgr object instance from fortimanager.py
    :type fmgr: class object
    :param paramgram: The formatted dictionary of options to process
    :type paramgram: dict
    :return: The response from the FortiManager
    :rtype: dict
    """
    # INIT A BASIC OBJECTS
    response = DEFAULT_RESULT_OBJ
    url = ""
    datagram = {}

    if paramgram["fmgr_ha_cluster_pw"] is not None and str(paramgram["fmgr_ha_mode"].lower()) != "standalone":
        datagram = {
            "mode": paramgram["fmgr_ha_mode"],
            "file-quota": paramgram["fmgr_ha_file_quota"],
            "hb-interval": paramgram["fmgr_ha_hb_interval"],
            "hb-lost-threshold": paramgram["fmgr_ha_hb_threshold"],
            "password": paramgram["fmgr_ha_cluster_pw"],
            "clusterid": paramgram["fmgr_ha_cluster_id"]
        }
    elif str(paramgram["fmgr_ha_mode"].lower()) == "standalone":
        datagram = {
            "mode": paramgram["fmgr_ha_mode"],
            "file-quota": paramgram["fmgr_ha_file_quota"],
            "hb-interval": paramgram["fmgr_ha_hb_interval"],
            "hb-lost-threshold": paramgram["fmgr_ha_hb_threshold"],
            "clusterid": paramgram["fmgr_ha_cluster_id"]
        }

    url = '/cli/global/system/ha'
    response = fmgr.process_request(url, datagram, FMGRMethods.SET)
    return response


def fmgr_get_ha_peer_list(fmgr):
    """
    :param fmgr: The fmgr object instance from fortimanager.py
    :type fmgr: class object
    :param paramgram: The formatted dictionary of options to process
    :type paramgram: dict
    :return: The response from the FortiManager
    :rtype: dict
    """
    # INIT A BASIC OBJECTS
    response = DEFAULT_RESULT_OBJ

    datagram = {}
    paramgram = {}

    url = '/cli/global/system/ha/peer/'
    response = fmgr.process_request(url, datagram, FMGRMethods.GET)
    return response


def fmgr_set_ha_peer(fmgr, paramgram):
    """
    :param fmgr: The fmgr object instance from fortimanager.py
    :type fmgr: class object
    :param paramgram: The formatted dictionary of options to process
    :type paramgram: dict
    :return: The response from the FortiManager
    :rtype: dict
    """

    datagram = {
        "ip": paramgram["fmgr_ha_peer_ipv4"],
        "ip6": paramgram["fmgr_ha_peer_ipv6"],
        "serial-number": paramgram["fmgr_ha_peer_sn"],
        "status": paramgram["fmgr_ha_peer_status"],
        "id": paramgram["peer_id"]
    }

    url = '/cli/global/system/ha/peer/'
    response = fmgr.process_request(url, datagram, FMGRMethods.SET)
    return response


def main():
    argument_spec = dict(
        fmgr_ha_mode=dict(required=False, type="str", choices=["standalone", "master", "slave"]),
        fmgr_ha_cluster_pw=dict(required=False, type="str", no_log=True),
        fmgr_ha_peer_status=dict(required=False, type="str", choices=["enable", "disable"]),
        fmgr_ha_peer_sn=dict(required=False, type="str"),
        fmgr_ha_peer_ipv4=dict(required=False, type="str"),
        fmgr_ha_peer_ipv6=dict(required=False, type="str"),
        fmgr_ha_hb_threshold=dict(required=False, type="int", default=3),
        fmgr_ha_hb_interval=dict(required=False, type="int", default=5),
        fmgr_ha_file_quota=dict(required=False, type="int", default=4096),
        fmgr_ha_cluster_id=dict(required=False, type="int", default=1)
    )

    required_if = [
        ['fmgr_ha_peer_ipv4', 'present', ['fmgr_ha_peer_sn', 'fmgr_ha_peer_status']],
        ['fmgr_ha_peer_ipv6', 'present', ['fmgr_ha_peer_sn', 'fmgr_ha_peer_status']],
        ['fmgr_ha_mode', 'master', ['fmgr_ha_cluster_pw', 'fmgr_ha_cluster_id']],
        ['fmgr_ha_mode', 'slave', ['fmgr_ha_cluster_pw', 'fmgr_ha_cluster_id']],
    ]

    module = AnsibleModule(argument_spec=argument_spec, supports_check_mode=False, required_if=required_if)
    paramgram = {
        "fmgr_ha_mode": module.params["fmgr_ha_mode"],
        "fmgr_ha_cluster_pw": module.params["fmgr_ha_cluster_pw"],
        "fmgr_ha_peer_status": module.params["fmgr_ha_peer_status"],
        "fmgr_ha_peer_sn": module.params["fmgr_ha_peer_sn"],
        "fmgr_ha_peer_ipv4": module.params["fmgr_ha_peer_ipv4"],
        "fmgr_ha_peer_ipv6": module.params["fmgr_ha_peer_ipv6"],
        "fmgr_ha_hb_threshold": module.params["fmgr_ha_hb_threshold"],
        "fmgr_ha_hb_interval": module.params["fmgr_ha_hb_interval"],
        "fmgr_ha_file_quota": module.params["fmgr_ha_file_quota"],
        "fmgr_ha_cluster_id": module.params["fmgr_ha_cluster_id"],
    }
    module.paramgram = paramgram
    fmgr = None
    if module._socket_path:
        connection = Connection(module._socket_path)
        fmgr = FortiManagerHandler(connection, module)
        fmgr.tools = FMGRCommon()
    else:
        module.fail_json(**FAIL_SOCKET_MSG)

    # INIT FLAGS AND COUNTERS
    get_ha_peers = 0
    results = DEFAULT_RESULT_OBJ
    try:
        if any(v is not None for v in (paramgram["fmgr_ha_peer_sn"], paramgram["fmgr_ha_peer_ipv4"],
                                       paramgram["fmgr_ha_peer_ipv6"], paramgram["fmgr_ha_peer_status"])):
            get_ha_peers = 1
    except Exception as err:
        raise FMGBaseException(err)
    try:
        # IF HA MODE IS NOT NULL, SWITCH THAT
        if paramgram["fmgr_ha_mode"] is not None:
            if (str.lower(paramgram["fmgr_ha_mode"]) != "standalone" and paramgram["fmgr_ha_cluster_pw"] is not None)\
                    or str.lower(paramgram["fmgr_ha_mode"]) == "standalone":
                results = fmgr_set_ha_mode(fmgr, paramgram)
                fmgr.govern_response(module=module, results=results, stop_on_success=False,
                                     ansible_facts=fmgr.construct_ansible_facts(results, module.params, paramgram))

            elif str.lower(paramgram["fmgr_ha_mode"]) != "standalone" and\
                    paramgram["fmgr_ha_mode"] is not None and\
                    paramgram["fmgr_ha_cluster_pw"] is None:
                module.exit_json(msg="If setting HA Mode of MASTER or SLAVE, you must specify a cluster password")

    except Exception as err:
        raise FMGBaseException(err)
        # IF GET_HA_PEERS IS ENABLED, LETS PROCESS THE PEERS
    try:
        if get_ha_peers == 1:
            # GET THE CURRENT LIST OF PEERS FROM THE NODE
            peers = fmgr_get_ha_peer_list(fmgr)
            # GET LENGTH OF RETURNED PEERS LIST AND ADD ONE FOR THE NEXT ID
            paramgram["next_peer_id"] = len(peers[1]) + 1
            # SET THE ACTUAL NUMBER OF PEERS
            num_of_peers = len(peers[1])
            # SET THE PEER ID FOR DISABLE METHOD
            paramgram["peer_id"] = len(peers) - 1
            # SET THE PEER LOOPCOUNT TO 1 TO START THE LOOP
            peer_loopcount = 1

            # LOOP THROUGH PEERS TO FIND THE SERIAL NUMBER MATCH TO GET THE RIGHT PEER ID
            # IDEA BEING WE DON'T WANT TO SUBMIT A BAD peer_id THAT DOESN'T JIVE WITH CURRENT DB ON FMG
            # SO LETS SEARCH FOR IT, AND IF WE FIND IT, WE WILL CHANGE THE PEER ID VARIABLES TO MATCH
            # IF NOT FOUND, LIFE GOES ON AND WE ASSUME THAT WE'RE ADDING A PEER
            # AT WHICH POINT THE next_peer_id VARIABLE WILL HAVE THE RIGHT PRIMARY KEY

            if paramgram["fmgr_ha_peer_sn"] is not None:
                while peer_loopcount <= num_of_peers:
                    # GET THE SERIAL NUMBER FOR CURRENT PEER IN LOOP TO COMPARE TO SN IN PLAYBOOK
                    try:
                        sn_compare = peers[1][peer_loopcount - 1]["serial-number"]
                        # IF THE SN IN THE PEERS MATCHES THE PLAYBOOK SN, SET THE IDS
                        if sn_compare == paramgram["fmgr_ha_peer_sn"]:
                            paramgram["peer_id"] = peer_loopcount
                            paramgram["next_peer_id"] = paramgram["peer_id"]
                    except Exception as err:
                        raise FMGBaseException(err)
                    # ADVANCE THE LOOP AND REPEAT UNTIL DONE
                    peer_loopcount += 1

            # IF THE PEER STATUS ISN'T IN THE PLAYBOOK, ASSUME ITS ENABLE
            if paramgram["fmgr_ha_peer_status"] is None:
                paramgram["fmgr_ha_peer_status"] = "enable"

            # IF THE PEER STATUS IS ENABLE, USE THE next_peer_id IN THE API CALL FOR THE ID
            if paramgram["fmgr_ha_peer_status"] == "enable":
                results = fmgr_set_ha_peer(fmgr, paramgram)
                fmgr.govern_response(module=module, results=results, stop_on_success=True,
                                     ansible_facts=fmgr.construct_ansible_facts(results,
                                                                                module.params, paramgram))

            # IF THE PEER STATUS IS DISABLE, WE HAVE TO HANDLE THAT A BIT DIFFERENTLY
            # JUST USING TWO DIFFERENT peer_id 's HERE
            if paramgram["fmgr_ha_peer_status"] == "disable":
                results = fmgr_set_ha_peer(fmgr, paramgram)
                fmgr.govern_response(module=module, results=results, stop_on_success=True,
                                     ansible_facts=fmgr.construct_ansible_facts(results, module.params, paramgram))

    except Exception as err:
        raise FMGBaseException(err)

    return module.exit_json(**results[1])


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Ansible Project
# Copyright: (c) 2019, Tobias Birkefeld (@tcraxs) <t@craxs.de>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = r'''
---
module: postgresql_privs
short_description: Grant or revoke privileges on PostgreSQL database objects
description:
- Grant or revoke privileges on PostgreSQL database objects.
- This module is basically a wrapper around most of the functionality of
  PostgreSQL's GRANT and REVOKE statements with detection of changes
  (GRANT/REVOKE I(privs) ON I(type) I(objs) TO/FROM I(roles)).
options:
  database:
    description:
    - Name of database to connect to.
    required: yes
    type: str
    aliases:
    - db
    - login_db
  state:
    description:
    - If C(present), the specified privileges are granted, if C(absent) they are revoked.
    type: str
    default: present
    choices: [ absent, present ]
  privs:
    description:
    - Comma separated list of privileges to grant/revoke.
    type: str
    aliases:
    - priv
  type:
    description:
    - Type of database object to set privileges on.
    - The C(default_privs) choice is available starting at version 2.7.
    - The C(foreign_data_wrapper) and C(foreign_server) object types are available since Ansible version 2.8.
    - The C(type) choice is available since Ansible version 2.10.
    - The C(procedure) is supported since collection version 1.3.0 and PostgreSQL 11.
    type: str
    default: table
    choices: [ database, default_privs, foreign_data_wrapper, foreign_server, function,
               group, language, table, tablespace, schema, sequence, type , procedure]
  objs:
    description:
    - Comma separated list of database objects to set privileges on.
    - If I(type) is C(table), C(partition table), C(sequence), C(function) or C(procedure),
      the special valueC(ALL_IN_SCHEMA) can be provided instead to specify all
      database objects of type I(type) in the schema specified via I(schema).
      (This also works with PostgreSQL < 9.0.) (C(ALL_IN_SCHEMA) is available
       for C(function) and C(partition table) since Ansible 2.8)
    - C(procedure) is supported since PostgreSQL 11 and M(community.general) collection 1.3.0.
    - If I(type) is C(database), this parameter can be omitted, in which case
      privileges are set for the database specified via I(database).
    - If I(type) is I(function) or I(procedure), colons (":") in object names will be
      replaced with commas (needed to specify signatures, see examples).
    type: str
    aliases:
    - obj
  schema:
    description:
    - Schema that contains the database objects specified via I(objs).
    - May only be provided if I(type) is C(table), C(sequence), C(function), C(procedure), C(type),
      or C(default_privs). Defaults to C(public) in these cases.
    - Pay attention, for embedded types when I(type=type)
      I(schema) can be C(pg_catalog) or C(information_schema) respectively.
    type: str
  roles:
    description:
    - Comma separated list of role (user/group) names to set permissions for.
    - The special value C(PUBLIC) can be provided instead to set permissions
      for the implicitly defined PUBLIC group.
    type: str
    required: yes
    aliases:
    - role
  fail_on_role:
    description:
    - If C(yes), fail when target role (for whom privs need to be granted) does not exist.
      Otherwise just warn and continue.
    default: yes
    type: bool
  session_role:
    description:
    - Switch to session_role after connecting.
    - The specified session_role must be a role that the current login_user is a member of.
    - Permissions checking for SQL commands is carried out as though the session_role were the one that had logged in originally.
    type: str
  target_roles:
    description:
    - A list of existing role (user/group) names to set as the
      default permissions for database objects subsequently created by them.
    - Parameter I(target_roles) is only available with C(type=default_privs).
    type: str
  grant_option:
    description:
    - Whether C(role) may grant/revoke the specified privileges/group memberships to others.
    - Set to C(no) to revoke GRANT OPTION, leave unspecified to make no changes.
    - I(grant_option) only has an effect if I(state) is C(present).
    type: bool
    aliases:
    - admin_option
  host:
    description:
    - Database host address. If unspecified, connect via Unix socket.
    type: str
    aliases:
    - login_host
  port:
    description:
    - Database port to connect to.
    type: int
    default: 5432
    aliases:
    - login_port
  unix_socket:
    description:
    - Path to a Unix domain socket for local connections.
    type: str
    aliases:
    - login_unix_socket
  login:
    description:
    - The username to authenticate with.
    type: str
    default: postgres
    aliases:
    - login_user
  password:
    description:
    - The password to authenticate with.
    type: str
    aliases:
    - login_password
  ssl_mode:
    description:
    - Determines whether or with what priority a secure SSL TCP/IP connection will be negotiated with the server.
    - See U(https://www.postgresql.org/docs/current/static/libpq-ssl.html) for more information on the modes.
    - Default of C(prefer) matches libpq default.
    type: str
    default: prefer
    choices: [ allow, disable, prefer, require, verify-ca, verify-full ]
  ca_cert:
    description:
    - Specifies the name of a file containing SSL certificate authority (CA) certificate(s).
    - If the file exists, the server's certificate will be verified to be signed by one of these authorities.
    type: str
    aliases:
    - ssl_rootcert
  trust_input:
    description:
    - If C(no), check whether values of parameters I(roles), I(target_roles), I(session_role),
      I(schema) are potentially dangerous.
    - It makes sense to use C(no) only when SQL injections via the parameters are possible.
    type: bool
    default: yes
    version_added: '0.2.0'
  usage_on_types:
    description:
    - When adding default privileges, the module always implicitly adds ``USAGE ON TYPES``.
    - To avoid this behavior, set I(usage_on_types) to C(no).
    - Added to save backwards compatibility.
    - Used only when adding default privileges, ignored otherwise.
    type: bool
    default: yes
    version_added: '1.2.0'

notes:
- Parameters that accept comma separated lists (I(privs), I(objs), I(roles))
  have singular alias names (I(priv), I(obj), I(role)).
- To revoke only C(GRANT OPTION) for a specific object, set I(state) to
  C(present) and I(grant_option) to C(no) (see examples).
- Note that when revoking privileges from a role R, this role  may still have
  access via privileges granted to any role R is a member of including C(PUBLIC).
- Note that when you use C(PUBLIC) role, the module always reports that the state has been changed.
- Note that when revoking privileges from a role R, you do so as the user
  specified via I(login). If R has been granted the same privileges by
  another user also, R can still access database objects via these privileges.
- When revoking privileges, C(RESTRICT) is assumed (see PostgreSQL docs).

seealso:
- module: community.general.postgresql_user
- module: community.general.postgresql_owner
- module: community.general.postgresql_membership
- name: PostgreSQL privileges
  description: General information about PostgreSQL privileges.
  link: https://www.postgresql.org/docs/current/ddl-priv.html
- name: PostgreSQL GRANT command reference
  description: Complete reference of the PostgreSQL GRANT command documentation.
  link: https://www.postgresql.org/docs/current/sql-grant.html
- name: PostgreSQL REVOKE command reference
  description: Complete reference of the PostgreSQL REVOKE command documentation.
  link: https://www.postgresql.org/docs/current/sql-revoke.html

extends_documentation_fragment:
- community.general.postgres


author:
- Bernhard Weitzhofer (@b6d)
- Tobias Birkefeld (@tcraxs)
'''

EXAMPLES = r'''
# On database "library":
# GRANT SELECT, INSERT, UPDATE ON TABLE public.books, public.authors
# TO librarian, reader WITH GRANT OPTION
- name: Grant privs to librarian and reader on database library
  community.general.postgresql_privs:
    database: library
    state: present
    privs: SELECT,INSERT,UPDATE
    type: table
    objs: books,authors
    schema: public
    roles: librarian,reader
    grant_option: yes

- name: Same as above leveraging default values
  community.general.postgresql_privs:
    db: library
    privs: SELECT,INSERT,UPDATE
    objs: books,authors
    roles: librarian,reader
    grant_option: yes

# REVOKE GRANT OPTION FOR INSERT ON TABLE books FROM reader
# Note that role "reader" will be *granted* INSERT privilege itself if this
# isn't already the case (since state: present).
- name: Revoke privs from reader
  community.general.postgresql_privs:
    db: library
    state: present
    priv: INSERT
    obj: books
    role: reader
    grant_option: no

# "public" is the default schema. This also works for PostgreSQL 8.x.
- name: REVOKE INSERT, UPDATE ON ALL TABLES IN SCHEMA public FROM reader
  community.general.postgresql_privs:
    db: library
    state: absent
    privs: INSERT,UPDATE
    objs: ALL_IN_SCHEMA
    role: reader

- name: GRANT ALL PRIVILEGES ON SCHEMA public, math TO librarian
  community.general.postgresql_privs:
    db: library
    privs: ALL
    type: schema
    objs: public,math
    role: librarian

# Note the separation of arguments with colons.
- name: GRANT ALL PRIVILEGES ON FUNCTION math.add(int, int) TO librarian, reader
  community.general.postgresql_privs:
    db: library
    privs: ALL
    type: function
    obj: add(int:int)
    schema: math
    roles: librarian,reader

# Note that group role memberships apply cluster-wide and therefore are not
# restricted to database "library" here.
- name: GRANT librarian, reader TO alice, bob WITH ADMIN OPTION
  community.general.postgresql_privs:
    db: library
    type: group
    objs: librarian,reader
    roles: alice,bob
    admin_option: yes

# Note that here "db: postgres" specifies the database to connect to, not the
# database to grant privileges on (which is specified via the "objs" param)
- name: GRANT ALL PRIVILEGES ON DATABASE library TO librarian
  community.general.postgresql_privs:
    db: postgres
    privs: ALL
    type: database
    obj: library
    role: librarian

# If objs is omitted for type "database", it defaults to the database
# to which the connection is established
- name: GRANT ALL PRIVILEGES ON DATABASE library TO librarian
  community.general.postgresql_privs:
    db: library
    privs: ALL
    type: database
    role: librarian

# Available since version 2.7
# Objs must be set, ALL_DEFAULT to TABLES/SEQUENCES/TYPES/FUNCTIONS
# ALL_DEFAULT works only with privs=ALL
# For specific
- name: ALTER DEFAULT PRIVILEGES ON DATABASE library TO librarian
  community.general.postgresql_privs:
    db: library
    objs: ALL_DEFAULT
    privs: ALL
    type: default_privs
    role: librarian
    grant_option: yes

# Available since version 2.7
# Objs must be set, ALL_DEFAULT to TABLES/SEQUENCES/TYPES/FUNCTIONS
# ALL_DEFAULT works only with privs=ALL
# For specific
- name: ALTER DEFAULT PRIVILEGES ON DATABASE library TO reader, step 1
  community.general.postgresql_privs:
    db: library
    objs: TABLES,SEQUENCES
    privs: SELECT
    type: default_privs
    role: reader

- name: ALTER DEFAULT PRIVILEGES ON DATABASE library TO reader, step 2
  community.general.postgresql_privs:
    db: library
    objs: TYPES
    privs: USAGE
    type: default_privs
    role: reader

# Available since version 2.8
- name: GRANT ALL PRIVILEGES ON FOREIGN DATA WRAPPER fdw TO reader
  community.general.postgresql_privs:
    db: test
    objs: fdw
    privs: ALL
    type: foreign_data_wrapper
    role: reader

# Available since community.general 0.2.0
- name: GRANT ALL PRIVILEGES ON TYPE customtype TO reader
  community.general.postgresql_privs:
    db: test
    objs: customtype
    privs: ALL
    type: type
    role: reader

# Available since version 2.8
- name: GRANT ALL PRIVILEGES ON FOREIGN SERVER fdw_server TO reader
  community.general.postgresql_privs:
    db: test
    objs: fdw_server
    privs: ALL
    type: foreign_server
    role: reader

# Available since version 2.8
# Grant 'execute' permissions on all functions in schema 'common' to role 'caller'
- name: GRANT EXECUTE ON ALL FUNCTIONS IN SCHEMA common TO caller
  community.general.postgresql_privs:
    type: function
    state: present
    privs: EXECUTE
    roles: caller
    objs: ALL_IN_SCHEMA
    schema: common

# Available since collection version 1.3.0
# Grant 'execute' permissions on all procedures in schema 'common' to role 'caller'
# Needs PostreSQL 11 or higher and community.general 1.3.0 or higher
- name: GRANT EXECUTE ON ALL PROCEDURES IN SCHEMA common TO caller
  community.general.postgresql_privs:
    type: prucedure
    state: present
    privs: EXECUTE
    roles: caller
    objs: ALL_IN_SCHEMA
    schema: common

# Available since version 2.8
# ALTER DEFAULT PRIVILEGES FOR ROLE librarian IN SCHEMA library GRANT SELECT ON TABLES TO reader
# GRANT SELECT privileges for new TABLES objects created by librarian as
# default to the role reader.
# For specific
- name: ALTER privs
  community.general.postgresql_privs:
    db: library
    schema: library
    objs: TABLES
    privs: SELECT
    type: default_privs
    role: reader
    target_roles: librarian

# Available since version 2.8
# ALTER DEFAULT PRIVILEGES FOR ROLE librarian IN SCHEMA library REVOKE SELECT ON TABLES FROM reader
# REVOKE SELECT privileges for new TABLES objects created by librarian as
# default from the role reader.
# For specific
- name: ALTER privs
  community.general.postgresql_privs:
    db: library
    state: absent
    schema: library
    objs: TABLES
    privs: SELECT
    type: default_privs
    role: reader
    target_roles: librarian

# Available since community.general 0.2.0
- name: Grant type privileges for pg_catalog.numeric type to alice
  community.general.postgresql_privs:
    type: type
    roles: alice
    privs: ALL
    objs: numeric
    schema: pg_catalog
    db: acme
'''

RETURN = r'''
queries:
  description: List of executed queries.
  returned: always
  type: list
  sample: ['REVOKE GRANT OPTION FOR INSERT ON TABLE "books" FROM "reader";']
'''

import traceback

PSYCOPG2_IMP_ERR = None
try:
    import psycopg2
    import psycopg2.extensions
except ImportError:
    PSYCOPG2_IMP_ERR = traceback.format_exc()
    psycopg2 = None

# import module snippets
from ansible.module_utils.basic import AnsibleModule, missing_required_lib
from ansible_collections.community.general.plugins.module_utils.database import (
    pg_quote_identifier,
    check_input,
)
from ansible_collections.community.general.plugins.module_utils.postgres import postgres_common_argument_spec
from ansible.module_utils._text import to_native

VALID_PRIVS = frozenset(('SELECT', 'INSERT', 'UPDATE', 'DELETE', 'TRUNCATE',
                         'REFERENCES', 'TRIGGER', 'CREATE', 'CONNECT',
                         'TEMPORARY', 'TEMP', 'EXECUTE', 'USAGE', 'ALL', 'USAGE'))
VALID_DEFAULT_OBJS = {'TABLES': ('ALL', 'SELECT', 'INSERT', 'UPDATE', 'DELETE', 'TRUNCATE', 'REFERENCES', 'TRIGGER'),
                      'SEQUENCES': ('ALL', 'SELECT', 'UPDATE', 'USAGE'),
                      'FUNCTIONS': ('ALL', 'EXECUTE'),
                      'TYPES': ('ALL', 'USAGE')}

executed_queries = []


class Error(Exception):
    pass


def role_exists(module, cursor, rolname):
    """Check user exists or not"""
    query = "SELECT 1 FROM pg_roles WHERE rolname = '%s'" % rolname
    try:
        cursor.execute(query)
        return cursor.rowcount > 0

    except Exception as e:
        module.fail_json(msg="Cannot execute SQL '%s': %s" % (query, to_native(e)))

    return False


# We don't have functools.partial in Python < 2.5
def partial(f, *args, **kwargs):
    """Partial function application"""

    def g(*g_args, **g_kwargs):
        new_kwargs = kwargs.copy()
        new_kwargs.update(g_kwargs)
        return f(*(args + g_args), **g_kwargs)

    g.f = f
    g.args = args
    g.kwargs = kwargs
    return g


class Connection(object):
    """Wrapper around a psycopg2 connection with some convenience methods"""

    def __init__(self, params, module):
        self.database = params.database
        self.module = module
        # To use defaults values, keyword arguments must be absent, so
        # check which values are empty and don't include in the **kw
        # dictionary
        params_map = {
            "host": "host",
            "login": "user",
            "password": "password",
            "port": "port",
            "database": "database",
            "ssl_mode": "sslmode",
            "ca_cert": "sslrootcert"
        }

        kw = dict((params_map[k], getattr(params, k)) for k in params_map
                  if getattr(params, k) != '' and getattr(params, k) is not None)

        # If a unix_socket is specified, incorporate it here.
        is_localhost = "host" not in kw or kw["host"] == "" or kw["host"] == "localhost"
        if is_localhost and params.unix_socket != "":
            kw["host"] = params.unix_socket

        sslrootcert = params.ca_cert
        if psycopg2.__version__ < '2.4.3' and sslrootcert is not None:
            raise ValueError('psycopg2 must be at least 2.4.3 in order to user the ca_cert parameter')

        self.connection = psycopg2.connect(**kw)
        self.cursor = self.connection.cursor()
        self.pg_version = self.connection.server_version

    def commit(self):
        self.connection.commit()

    def rollback(self):
        self.connection.rollback()

    @property
    def encoding(self):
        """Connection encoding in Python-compatible form"""
        return psycopg2.extensions.encodings[self.connection.encoding]

    # Methods for querying database objects

    # PostgreSQL < 9.0 doesn't support "ALL TABLES IN SCHEMA schema"-like
    # phrases in GRANT or REVOKE statements, therefore alternative methods are
    # provided here.

    def schema_exists(self, schema):
        query = """SELECT count(*)
                   FROM pg_catalog.pg_namespace WHERE nspname = %s"""
        self.cursor.execute(query, (schema,))
        return self.cursor.fetchone()[0] > 0

    def get_all_tables_in_schema(self, schema):
        if not self.schema_exists(schema):
            raise Error('Schema "%s" does not exist.' % schema)
        query = """SELECT relname
                   FROM pg_catalog.pg_class c
                   JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace
                   WHERE nspname = %s AND relkind in ('r', 'v', 'm', 'p')"""
        self.cursor.execute(query, (schema,))
        return [t[0] for t in self.cursor.fetchall()]

    def get_all_sequences_in_schema(self, schema):
        if not self.schema_exists(schema):
            raise Error('Schema "%s" does not exist.' % schema)
        query = """SELECT relname
                   FROM pg_catalog.pg_class c
                   JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace
                   WHERE nspname = %s AND relkind = 'S'"""
        self.cursor.execute(query, (schema,))
        return [t[0] for t in self.cursor.fetchall()]

    def get_all_functions_in_schema(self, schema):
        if not self.schema_exists(schema):
            raise Error('Schema "%s" does not exist.' % schema)

        query = ("SELECT p.proname, oidvectortypes(p.proargtypes) "
                 "FROM pg_catalog.pg_proc p "
                 "JOIN pg_namespace n ON n.oid = p.pronamespace "
                 "WHERE nspname = %s")

        if self.pg_version >= 110000:
            query += " and p.prokind = 'f'"

        self.cursor.execute(query, (schema,))
        return ["%s(%s)" % (t[0], t[1]) for t in self.cursor.fetchall()]

    def get_all_procedures_in_schema(self, schema):
        if self.pg_version < 110000:
            raise Error("PostgreSQL verion must be >= 11 for type=procedure. Exit")

        if not self.schema_exists(schema):
            raise Error('Schema "%s" does not exist.' % schema)

        query = ("SELECT p.proname, oidvectortypes(p.proargtypes) "
                 "FROM pg_catalog.pg_proc p "
                 "JOIN pg_namespace n ON n.oid = p.pronamespace "
                 "WHERE nspname = %s and p.prokind = 'p'")

        self.cursor.execute(query, (schema,))
        return ["%s(%s)" % (t[0], t[1]) for t in self.cursor.fetchall()]

    # Methods for getting access control lists and group membership info

    # To determine whether anything has changed after granting/revoking
    # privileges, we compare the access control lists of the specified database
    # objects before and afterwards. Python's list/string comparison should
    # suffice for change detection, we should not actually have to parse ACLs.
    # The same should apply to group membership information.

    def get_table_acls(self, schema, tables):
        query = """SELECT relacl
                   FROM pg_catalog.pg_class c
                   JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace
                   WHERE nspname = %s AND relkind in ('r','p','v','m') AND relname = ANY (%s)
                   ORDER BY relname"""
        self.cursor.execute(query, (schema, tables))
        return [t[0] for t in self.cursor.fetchall()]

    def get_sequence_acls(self, schema, sequences):
        query = """SELECT relacl
                   FROM pg_catalog.pg_class c
                   JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace
                   WHERE nspname = %s AND relkind = 'S' AND relname = ANY (%s)
                   ORDER BY relname"""
        self.cursor.execute(query, (schema, sequences))
        return [t[0] for t in self.cursor.fetchall()]

    def get_function_acls(self, schema, function_signatures):
        funcnames = [f.split('(', 1)[0] for f in function_signatures]
        query = """SELECT proacl
                   FROM pg_catalog.pg_proc p
                   JOIN pg_catalog.pg_namespace n ON n.oid = p.pronamespace
                   WHERE nspname = %s AND proname = ANY (%s)
                   ORDER BY proname, proargtypes"""
        self.cursor.execute(query, (schema, funcnames))
        return [t[0] for t in self.cursor.fetchall()]

    def get_schema_acls(self, schemas):
        query = """SELECT nspacl FROM pg_catalog.pg_namespace
                   WHERE nspname = ANY (%s) ORDER BY nspname"""
        self.cursor.execute(query, (schemas,))
        return [t[0] for t in self.cursor.fetchall()]

    def get_language_acls(self, languages):
        query = """SELECT lanacl FROM pg_catalog.pg_language
                   WHERE lanname = ANY (%s) ORDER BY lanname"""
        self.cursor.execute(query, (languages,))
        return [t[0] for t in self.cursor.fetchall()]

    def get_tablespace_acls(self, tablespaces):
        query = """SELECT spcacl FROM pg_catalog.pg_tablespace
                   WHERE spcname = ANY (%s) ORDER BY spcname"""
        self.cursor.execute(query, (tablespaces,))
        return [t[0] for t in self.cursor.fetchall()]

    def get_database_acls(self, databases):
        query = """SELECT datacl FROM pg_catalog.pg_database
                   WHERE datname = ANY (%s) ORDER BY datname"""
        self.cursor.execute(query, (databases,))
        return [t[0] for t in self.cursor.fetchall()]

    def get_group_memberships(self, groups):
        query = """SELECT roleid, grantor, member, admin_option
                   FROM pg_catalog.pg_auth_members am
                   JOIN pg_catalog.pg_roles r ON r.oid = am.roleid
                   WHERE r.rolname = ANY(%s)
                   ORDER BY roleid, grantor, member"""
        self.cursor.execute(query, (groups,))
        return self.cursor.fetchall()

    def get_default_privs(self, schema, *args):
        query = """SELECT defaclacl
                   FROM pg_default_acl a
                   JOIN pg_namespace b ON a.defaclnamespace=b.oid
                   WHERE b.nspname = %s;"""
        self.cursor.execute(query, (schema,))
        return [t[0] for t in self.cursor.fetchall()]

    def get_foreign_data_wrapper_acls(self, fdws):
        query = """SELECT fdwacl FROM pg_catalog.pg_foreign_data_wrapper
                   WHERE fdwname = ANY (%s) ORDER BY fdwname"""
        self.cursor.execute(query, (fdws,))
        return [t[0] for t in self.cursor.fetchall()]

    def get_foreign_server_acls(self, fs):
        query = """SELECT srvacl FROM pg_catalog.pg_foreign_server
                   WHERE srvname = ANY (%s) ORDER BY srvname"""
        self.cursor.execute(query, (fs,))
        return [t[0] for t in self.cursor.fetchall()]

    def get_type_acls(self, schema, types):
        query = """SELECT t.typacl FROM pg_catalog.pg_type t
                   JOIN pg_catalog.pg_namespace n ON n.oid = t.typnamespace
                   WHERE n.nspname = %s AND t.typname = ANY (%s) ORDER BY typname"""
        self.cursor.execute(query, (schema, types))
        return [t[0] for t in self.cursor.fetchall()]

    # Manipulating privileges

    def manipulate_privs(self, obj_type, privs, objs, roles, target_roles,
                         state, grant_option, schema_qualifier=None, fail_on_role=True, usage_on_types=True):
        """Manipulate database object privileges.

        :param obj_type: Type of database object to grant/revoke
                         privileges for.
        :param privs: Either a list of privileges to grant/revoke
                      or None if type is "group".
        :param objs: List of database objects to grant/revoke
                     privileges for.
        :param roles: Either a list of role names or "PUBLIC"
                      for the implicitly defined "PUBLIC" group
        :param target_roles: List of role names to grant/revoke
                             default privileges as.
        :param state: "present" to grant privileges, "absent" to revoke.
        :param grant_option: Only for state "present": If True, set
                             grant/admin option. If False, revoke it.
                             If None, don't change grant option.
        :param schema_qualifier: Some object types ("TABLE", "SEQUENCE",
                                 "FUNCTION") must be qualified by schema.
                                 Ignored for other Types.
        """
        # get_status: function to get current status
        if obj_type == 'table':
            get_status = partial(self.get_table_acls, schema_qualifier)
        elif obj_type == 'sequence':
            get_status = partial(self.get_sequence_acls, schema_qualifier)
        elif obj_type in ('function', 'procedure'):
            get_status = partial(self.get_function_acls, schema_qualifier)
        elif obj_type == 'schema':
            get_status = self.get_schema_acls
        elif obj_type == 'language':
            get_status = self.get_language_acls
        elif obj_type == 'tablespace':
            get_status = self.get_tablespace_acls
        elif obj_type == 'database':
            get_status = self.get_database_acls
        elif obj_type == 'group':
            get_status = self.get_group_memberships
        elif obj_type == 'default_privs':
            get_status = partial(self.get_default_privs, schema_qualifier)
        elif obj_type == 'foreign_data_wrapper':
            get_status = self.get_foreign_data_wrapper_acls
        elif obj_type == 'foreign_server':
            get_status = self.get_foreign_server_acls
        elif obj_type == 'type':
            get_status = partial(self.get_type_acls, schema_qualifier)
        else:
            raise Error('Unsupported database object type "%s".' % obj_type)

        # Return False (nothing has changed) if there are no objs to work on.
        if not objs:
            return False

        # obj_ids: quoted db object identifiers (sometimes schema-qualified)
        if obj_type in ('function', 'procedure'):
            obj_ids = []
            for obj in objs:
                try:
                    f, args = obj.split('(', 1)
                except Exception:
                    raise Error('Illegal function / procedure signature: "%s".' % obj)
                obj_ids.append('"%s"."%s"(%s' % (schema_qualifier, f, args))
        elif obj_type in ['table', 'sequence', 'type']:
            obj_ids = ['"%s"."%s"' % (schema_qualifier, o) for o in objs]
        else:
            obj_ids = ['"%s"' % o for o in objs]

        # set_what: SQL-fragment specifying what to set for the target roles:
        # Either group membership or privileges on objects of a certain type
        if obj_type == 'group':
            set_what = ','.join(obj_ids)
        elif obj_type == 'default_privs':
            # We don't want privs to be quoted here
            set_what = ','.join(privs)
        else:
            # function types are already quoted above
            if obj_type not in ('function', 'procedure'):
                obj_ids = [pg_quote_identifier(i, 'table') for i in obj_ids]
            # Note: obj_type has been checked against a set of string literals
            # and privs was escaped when it was parsed
            # Note: Underscores are replaced with spaces to support multi-word obj_type
            set_what = '%s ON %s %s' % (','.join(privs), obj_type.replace('_', ' '),
                                        ','.join(obj_ids))

        # for_whom: SQL-fragment specifying for whom to set the above
        if roles == 'PUBLIC':
            for_whom = 'PUBLIC'
        else:
            for_whom = []
            for r in roles:
                if not role_exists(self.module, self.cursor, r):
                    if fail_on_role:
                        self.module.fail_json(msg="Role '%s' does not exist" % r.strip())

                    else:
                        self.module.warn("Role '%s' does not exist, pass it" % r.strip())
                else:
                    for_whom.append('"%s"' % r)

            if not for_whom:
                return False

            for_whom = ','.join(for_whom)

        # as_who:
        as_who = None
        if target_roles:
            as_who = ','.join('"%s"' % r for r in target_roles)

        if schema_qualifier:
            schema_qualifier = '"%s"' % schema_qualifier

        status_before = get_status(objs)

        query = QueryBuilder(state) \
            .for_objtype(obj_type) \
            .with_grant_option(grant_option) \
            .for_whom(for_whom) \
            .as_who(as_who) \
            .for_schema(schema_qualifier) \
            .set_what(set_what) \
            .for_objs(objs) \
            .usage_on_types(usage_on_types) \
            .build()

        executed_queries.append(query)
        self.cursor.execute(query)
        if roles == 'PUBLIC':
            return True

        status_after = get_status(objs)

        def nonesorted(e):
            # For python 3+ that can fail trying
            # to compare NoneType elements by sort method.
            if e is None:
                return ''
            return e

        status_before.sort(key=nonesorted)
        status_after.sort(key=nonesorted)
        return status_before != status_after


class QueryBuilder(object):
    def __init__(self, state):
        self._grant_option = None
        self._for_whom = None
        self._as_who = None
        self._set_what = None
        self._obj_type = None
        self._state = state
        self._schema = None
        self._objs = None
        self._usage_on_types = None
        self.query = []

    def for_objs(self, objs):
        self._objs = objs
        return self

    def for_schema(self, schema):
        self._schema = schema
        return self

    def with_grant_option(self, option):
        self._grant_option = option
        return self

    def for_whom(self, who):
        self._for_whom = who
        return self

    def usage_on_types(self, usage_on_types):
        self._usage_on_types = usage_on_types
        return self

    def as_who(self, target_roles):
        self._as_who = target_roles
        return self

    def set_what(self, what):
        self._set_what = what
        return self

    def for_objtype(self, objtype):
        self._obj_type = objtype
        return self

    def build(self):
        if self._state == 'present':
            self.build_present()
        elif self._state == 'absent':
            self.build_absent()
        else:
            self.build_absent()
        return '\n'.join(self.query)

    def add_default_revoke(self):
        for obj in self._objs:
            if self._as_who:
                self.query.append(
                    'ALTER DEFAULT PRIVILEGES FOR ROLE {0} IN SCHEMA {1} REVOKE ALL ON {2} FROM {3};'.format(self._as_who,
                                                                                                             self._schema, obj,
                                                                                                             self._for_whom))
            else:
                self.query.append(
                    'ALTER DEFAULT PRIVILEGES IN SCHEMA {0} REVOKE ALL ON {1} FROM {2};'.format(self._schema, obj,
                                                                                                self._for_whom))

    def add_grant_option(self):
        if self._grant_option:
            if self._obj_type == 'group':
                self.query[-1] += ' WITH ADMIN OPTION;'
            else:
                self.query[-1] += ' WITH GRANT OPTION;'
        elif self._grant_option is False:
            self.query[-1] += ';'
            if self._obj_type == 'group':
                self.query.append('REVOKE ADMIN OPTION FOR {0} FROM {1};'.format(self._set_what, self._for_whom))
            elif not self._obj_type == 'default_privs':
                self.query.append('REVOKE GRANT OPTION FOR {0} FROM {1};'.format(self._set_what, self._for_whom))
        else:
            self.query[-1] += ';'

    def add_default_priv(self):
        for obj in self._objs:
            if self._as_who:
                self.query.append(
                    'ALTER DEFAULT PRIVILEGES FOR ROLE {0} IN SCHEMA {1} GRANT {2} ON {3} TO {4}'.format(self._as_who,
                                                                                                         self._schema,
                                                                                                         self._set_what,
                                                                                                         obj,
                                                                                                         self._for_whom))
            else:
                self.query.append(
                    'ALTER DEFAULT PRIVILEGES IN SCHEMA {0} GRANT {1} ON {2} TO {3}'.format(self._schema,
                                                                                            self._set_what,
                                                                                            obj,
                                                                                            self._for_whom))
            self.add_grant_option()

        if self._usage_on_types:
            if self._as_who:
                self.query.append(
                    'ALTER DEFAULT PRIVILEGES FOR ROLE {0} IN SCHEMA {1} GRANT USAGE ON TYPES TO {2}'.format(self._as_who,
                                                                                                             self._schema,
                                                                                                             self._for_whom))
            else:
                self.query.append(
                    'ALTER DEFAULT PRIVILEGES IN SCHEMA {0} GRANT USAGE ON TYPES TO {1}'.format(self._schema, self._for_whom))
        self.add_grant_option()

    def build_present(self):
        if self._obj_type == 'default_privs':
            self.add_default_revoke()
            self.add_default_priv()
        else:
            self.query.append('GRANT {0} TO {1}'.format(self._set_what, self._for_whom))
            self.add_grant_option()

    def build_absent(self):
        if self._obj_type == 'default_privs':
            self.query = []
            for obj in ['TABLES', 'SEQUENCES', 'TYPES']:
                if self._as_who:
                    self.query.append(
                        'ALTER DEFAULT PRIVILEGES FOR ROLE {0} IN SCHEMA {1} REVOKE ALL ON {2} FROM {3};'.format(self._as_who,
                                                                                                                 self._schema, obj,
                                                                                                                 self._for_whom))
                else:
                    self.query.append(
                        'ALTER DEFAULT PRIVILEGES IN SCHEMA {0} REVOKE ALL ON {1} FROM {2};'.format(self._schema, obj,
                                                                                                    self._for_whom))
        else:
            self.query.append('REVOKE {0} FROM {1};'.format(self._set_what, self._for_whom))


def main():
    argument_spec = postgres_common_argument_spec()
    argument_spec.update(
        database=dict(required=True, aliases=['db', 'login_db']),
        state=dict(default='present', choices=['present', 'absent']),
        privs=dict(required=False, aliases=['priv']),
        type=dict(default='table',
                  choices=['table',
                           'sequence',
                           'function',
                           'procedure',
                           'database',
                           'schema',
                           'language',
                           'tablespace',
                           'group',
                           'default_privs',
                           'foreign_data_wrapper',
                           'foreign_server',
                           'type', ]),
        objs=dict(required=False, aliases=['obj']),
        schema=dict(required=False),
        roles=dict(required=True, aliases=['role']),
        session_role=dict(required=False),
        target_roles=dict(required=False),
        grant_option=dict(required=False, type='bool',
                          aliases=['admin_option']),
        host=dict(default='', aliases=['login_host']),
        unix_socket=dict(default='', aliases=['login_unix_socket']),
        login=dict(default='postgres', aliases=['login_user']),
        password=dict(default='', aliases=['login_password'], no_log=True),
        fail_on_role=dict(type='bool', default=True),
        trust_input=dict(type='bool', default=True),
        usage_on_types=dict(type='bool', default=True),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )

    fail_on_role = module.params['fail_on_role']
    usage_on_types = module.params['usage_on_types']

    # Create type object as namespace for module params
    p = type('Params', (), module.params)
    # param "schema": default, allowed depends on param "type"
    if p.type in ['table', 'sequence', 'function', 'procedure', 'type', 'default_privs']:
        p.schema = p.schema or 'public'
    elif p.schema:
        module.fail_json(msg='Argument "schema" is not allowed '
                             'for type "%s".' % p.type)

    # param "objs": default, required depends on param "type"
    if p.type == 'database':
        p.objs = p.objs or p.database
    elif not p.objs:
        module.fail_json(msg='Argument "objs" is required '
                             'for type "%s".' % p.type)

    # param "privs": allowed, required depends on param "type"
    if p.type == 'group':
        if p.privs:
            module.fail_json(msg='Argument "privs" is not allowed '
                                 'for type "group".')
    elif not p.privs:
        module.fail_json(msg='Argument "privs" is required '
                             'for type "%s".' % p.type)

    # Check input
    if not p.trust_input:
        # Check input for potentially dangerous elements:
        check_input(module, p.roles, p.target_roles, p.session_role, p.schema)

    # Connect to Database
    if not psycopg2:
        module.fail_json(msg=missing_required_lib('psycopg2'), exception=PSYCOPG2_IMP_ERR)
    try:
        conn = Connection(p, module)
    except psycopg2.Error as e:
        module.fail_json(msg='Could not connect to database: %s' % to_native(e), exception=traceback.format_exc())
    except TypeError as e:
        if 'sslrootcert' in e.args[0]:
            module.fail_json(msg='Postgresql server must be at least version 8.4 to support sslrootcert')
        module.fail_json(msg="unable to connect to database: %s" % to_native(e), exception=traceback.format_exc())
    except ValueError as e:
        # We raise this when the psycopg library is too old
        module.fail_json(msg=to_native(e))

    if p.session_role:
        try:
            conn.cursor.execute('SET ROLE "%s"' % p.session_role)
        except Exception as e:
            module.fail_json(msg="Could not switch to role %s: %s" % (p.session_role, to_native(e)), exception=traceback.format_exc())

    try:
        # privs
        if p.privs:
            privs = frozenset(pr.upper() for pr in p.privs.split(','))
            if not privs.issubset(VALID_PRIVS):
                module.fail_json(msg='Invalid privileges specified: %s' % privs.difference(VALID_PRIVS))
        else:
            privs = None
        # objs:
        if p.type == 'table' and p.objs == 'ALL_IN_SCHEMA':
            objs = conn.get_all_tables_in_schema(p.schema)
        elif p.type == 'sequence' and p.objs == 'ALL_IN_SCHEMA':
            objs = conn.get_all_sequences_in_schema(p.schema)
        elif p.type == 'function' and p.objs == 'ALL_IN_SCHEMA':
            objs = conn.get_all_functions_in_schema(p.schema)
        elif p.type == 'procedure' and p.objs == 'ALL_IN_SCHEMA':
            objs = conn.get_all_procedures_in_schema(p.schema)
        elif p.type == 'default_privs':
            if p.objs == 'ALL_DEFAULT':
                objs = frozenset(VALID_DEFAULT_OBJS.keys())
            else:
                objs = frozenset(obj.upper() for obj in p.objs.split(','))
                if not objs.issubset(VALID_DEFAULT_OBJS):
                    module.fail_json(
                        msg='Invalid Object set specified: %s' % objs.difference(VALID_DEFAULT_OBJS.keys()))
            # Again, do we have valid privs specified for object type:
            valid_objects_for_priv = frozenset(obj for obj in objs if privs.issubset(VALID_DEFAULT_OBJS[obj]))
            if not valid_objects_for_priv == objs:
                module.fail_json(
                    msg='Invalid priv specified. Valid object for priv: {0}. Objects: {1}'.format(
                        valid_objects_for_priv, objs))
        else:
            objs = p.objs.split(',')

            # function signatures are encoded using ':' to separate args
            if p.type in ('function', 'procedure'):
                objs = [obj.replace(':', ',') for obj in objs]

        # roles
        if p.roles.upper() == 'PUBLIC':
            roles = 'PUBLIC'
        else:
            roles = p.roles.split(',')

            if len(roles) == 1 and not role_exists(module, conn.cursor, roles[0]):
                module.exit_json(changed=False)

                if fail_on_role:
                    module.fail_json(msg="Role '%s' does not exist" % roles[0].strip())

                else:
                    module.warn("Role '%s' does not exist, nothing to do" % roles[0].strip())

        # check if target_roles is set with type: default_privs
        if p.target_roles and not p.type == 'default_privs':
            module.warn('"target_roles" will be ignored '
                        'Argument "type: default_privs" is required for usage of "target_roles".')

        # target roles
        if p.target_roles:
            target_roles = p.target_roles.split(',')
        else:
            target_roles = None

        changed = conn.manipulate_privs(
            obj_type=p.type,
            privs=privs,
            objs=objs,
            roles=roles,
            target_roles=target_roles,
            state=p.state,
            grant_option=p.grant_option,
            schema_qualifier=p.schema,
            fail_on_role=fail_on_role,
            usage_on_types=usage_on_types,
        )

    except Error as e:
        conn.rollback()
        module.fail_json(msg=to_native(e), exception=traceback.format_exc())

    except psycopg2.Error as e:
        conn.rollback()
        module.fail_json(msg=to_native(e))

    if module.check_mode or not changed:
        conn.rollback()
    else:
        conn.commit()
    module.exit_json(changed=changed, queries=executed_queries)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2019, Tobias Birkefeld (@tcraxs) <t@craxs.de>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

DOCUMENTATION = r'''
---
module: postgresql_sequence
short_description: Create, drop, or alter a PostgreSQL sequence
description:
- Allows to create, drop or change the definition of a sequence generator.
options:
  sequence:
    description:
    - The name of the sequence.
    required: true
    type: str
    aliases:
    - name
  state:
    description:
    - The sequence state.
    - If I(state=absent) other options will be ignored except of I(name) and
      I(schema).
    default: present
    choices: [ absent, present ]
    type: str
  data_type:
    description:
    - Specifies the data type of the sequence. Valid types are bigint, integer,
      and smallint. bigint is the default. The data type determines the default
      minimum and maximum values of the sequence. For more info see the
      documentation
      U(https://www.postgresql.org/docs/current/sql-createsequence.html).
    - Supported from PostgreSQL 10.
    choices: [ bigint, integer, smallint ]
    type: str
  increment:
    description:
    - Increment specifies which value is added to the current sequence value
      to create a new value.
    - A positive value will make an ascending sequence, a negative one a
      descending sequence. The default value is 1.
    type: int
  minvalue:
    description:
    - Minvalue determines the minimum value a sequence can generate. The
      default for an ascending sequence is 1. The default for a descending
      sequence is the minimum value of the data type.
    type: int
    aliases:
      - min
  maxvalue:
    description:
    - Maxvalue determines the maximum value for the sequence. The default for
      an ascending sequence is the maximum
      value of the data type. The default for a descending sequence is -1.
    type: int
    aliases:
      - max
  start:
    description:
    - Start allows the sequence to begin anywhere. The default starting value
      is I(minvalue) for ascending sequences and I(maxvalue) for descending
      ones.
    type: int
  cache:
    description:
    - Cache specifies how many sequence numbers are to be preallocated and
      stored in memory for faster access. The minimum value is 1 (only one
      value can be generated at a time, i.e., no cache), and this is also
      the default.
    type: int
  cycle:
    description:
    - The cycle option allows the sequence to wrap around when the I(maxvalue)
      or I(minvalue) has been reached by an ascending or descending sequence
      respectively. If the limit is reached, the next number generated will be
      the minvalue or maxvalue, respectively.
    - If C(false) (NO CYCLE) is specified, any calls to nextval after the sequence
      has reached its maximum value will return an error. False (NO CYCLE) is
      the default.
    type: bool
    default: no
  cascade:
    description:
    - Automatically drop objects that depend on the sequence, and in turn all
      objects that depend on those objects.
    - Ignored if I(state=present).
    - Only used with I(state=absent).
    type: bool
    default: no
  rename_to:
    description:
    - The new name for the I(sequence).
    - Works only for existing sequences.
    type: str
  owner:
    description:
    - Set the owner for the I(sequence).
    type: str
  schema:
    description:
    - The schema of the I(sequence). This is be used to create and relocate
      a I(sequence) in the given schema.
    default: public
    type: str
  newschema:
    description:
    - The new schema for the I(sequence). Will be used for moving a
      I(sequence) to another I(schema).
    - Works only for existing sequences.
    type: str
  session_role:
    description:
    - Switch to session_role after connecting. The specified I(session_role)
      must be a role that the current I(login_user) is a member of.
    - Permissions checking for SQL commands is carried out as though
      the I(session_role) were the one that had logged in originally.
    type: str
  db:
    description:
    - Name of database to connect to and run queries against.
    type: str
    aliases:
    - database
    - login_db
  trust_input:
    description:
    - If C(no), check whether values of parameters I(sequence), I(schema), I(rename_to),
      I(owner), I(newschema), I(session_role) are potentially dangerous.
    - It makes sense to use C(no) only when SQL injections via the parameters are possible.
    type: bool
    default: yes
    version_added: '0.2.0'
notes:
- If you do not pass db parameter, sequence will be created in the database
  named postgres.
seealso:
- module: community.general.postgresql_table
- module: community.general.postgresql_owner
- module: community.general.postgresql_privs
- module: community.general.postgresql_tablespace
- name: CREATE SEQUENCE reference
  description: Complete reference of the CREATE SEQUENCE command documentation.
  link: https://www.postgresql.org/docs/current/sql-createsequence.html
- name: ALTER SEQUENCE reference
  description: Complete reference of the ALTER SEQUENCE command documentation.
  link: https://www.postgresql.org/docs/current/sql-altersequence.html
- name: DROP SEQUENCE reference
  description: Complete reference of the DROP SEQUENCE command documentation.
  link: https://www.postgresql.org/docs/current/sql-dropsequence.html
author:
- Tobias Birkefeld (@tcraxs)
- Thomas O'Donnell (@andytom)
extends_documentation_fragment:
- community.general.postgres

'''

EXAMPLES = r'''
- name: Create an ascending bigint sequence called foobar in the default
        database
  community.general.postgresql_sequence:
    name: foobar

- name: Create an ascending integer sequence called foobar, starting at 101
  community.general.postgresql_sequence:
    name: foobar
    data_type: integer
    start: 101

- name: Create an descending sequence called foobar, starting at 101 and
        preallocated 10 sequence numbers in cache
  community.general.postgresql_sequence:
    name: foobar
    increment: -1
    cache: 10
    start: 101

- name: Create an ascending sequence called foobar, which cycle between 1 to 10
  community.general.postgresql_sequence:
    name: foobar
    cycle: yes
    min: 1
    max: 10

- name: Create an ascending bigint sequence called foobar in the default
        database with owner foobar
  community.general.postgresql_sequence:
    name: foobar
    owner: foobar

- name: Rename an existing sequence named foo to bar
  community.general.postgresql_sequence:
    name: foo
    rename_to: bar

- name: Change the schema of an existing sequence to foobar
  community.general.postgresql_sequence:
    name: foobar
    newschema: foobar

- name: Change the owner of an existing sequence to foobar
  community.general.postgresql_sequence:
    name: foobar
    owner: foobar

- name: Drop a sequence called foobar
  community.general.postgresql_sequence:
    name: foobar
    state: absent

- name: Drop a sequence called foobar with cascade
  community.general.postgresql_sequence:
    name: foobar
    cascade: yes
    state: absent
'''

RETURN = r'''
state:
  description: Sequence state at the end of execution.
  returned: always
  type: str
  sample: 'present'
sequence:
  description: Sequence name.
  returned: always
  type: str
  sample: 'foobar'
queries:
    description: List of queries that was tried to be executed.
    returned: always
    type: str
    sample: [ "CREATE SEQUENCE \"foo\"" ]
schema:
    description: Name of the schema of the sequence
    returned: always
    type: str
    sample: 'foo'
data_type:
    description: Shows the current data type of the sequence.
    returned: always
    type: str
    sample: 'bigint'
increment:
    description: The value of increment of the sequence. A positive value will
                 make an ascending sequence, a negative one a descending
                 sequence.
    returned: always
    type: int
    sample: '-1'
minvalue:
    description: The value of minvalue of the sequence.
    returned: always
    type: int
    sample: '1'
maxvalue:
    description: The value of maxvalue of the sequence.
    returned: always
    type: int
    sample: '9223372036854775807'
start:
    description: The value of start of the sequence.
    returned: always
    type: int
    sample: '12'
cycle:
    description: Shows if the sequence cycle or not.
    returned: always
    type: str
    sample: 'NO'
owner:
    description: Shows the current owner of the sequence
                 after the successful run of the task.
    returned: always
    type: str
    sample: 'postgres'
newname:
    description: Shows the new sequence name after rename.
    returned: on success
    type: str
    sample: 'barfoo'
newschema:
    description: Shows the new schema of the sequence after schema change.
    returned: on success
    type: str
    sample: 'foobar'
'''


try:
    from psycopg2.extras import DictCursor
except ImportError:
    # psycopg2 is checked by connect_to_db()
    # from ansible.module_utils.postgres
    pass

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.community.general.plugins.module_utils.database import (
    check_input,
)
from ansible_collections.community.general.plugins.module_utils.postgres import (
    connect_to_db,
    exec_sql,
    get_conn_params,
    postgres_common_argument_spec,
)


class Sequence(object):
    """Implements behavior of CREATE, ALTER or DROP SEQUENCE PostgreSQL command.

    Arguments:
        module (AnsibleModule) -- object of AnsibleModule class
        cursor (cursor) -- cursor object of psycopg2 library

    Attributes:
        module (AnsibleModule) -- object of AnsibleModule class
        cursor (cursor) -- cursor object of psycopg2 library
        changed (bool) --  something was changed after execution or not
        executed_queries (list) -- executed queries
        name (str) -- name of the sequence
        owner (str) -- name of the owner of the sequence
        schema (str) -- name of the schema (default: public)
        data_type (str) -- data type of the sequence
        start_value (int) -- value of the sequence start
        minvalue (int) -- minimum value of the sequence
        maxvalue (int) -- maximum value of the sequence
        increment (int) -- increment value of the sequence
        cycle (bool) -- sequence can cycle or not
        new_name (str) -- name of the renamed sequence
        new_schema (str) -- name of the new schema
        exists (bool) -- sequence exists or not
    """

    def __init__(self, module, cursor):
        self.module = module
        self.cursor = cursor
        self.executed_queries = []
        self.name = self.module.params['sequence']
        self.owner = ''
        self.schema = self.module.params['schema']
        self.data_type = ''
        self.start_value = ''
        self.minvalue = ''
        self.maxvalue = ''
        self.increment = ''
        self.cycle = ''
        self.new_name = ''
        self.new_schema = ''
        self.exists = False
        # Collect info
        self.get_info()

    def get_info(self):
        """Getter to refresh and get sequence info"""
        query = ("SELECT "
                 "s.sequence_schema AS schemaname, "
                 "s.sequence_name AS sequencename, "
                 "pg_get_userbyid(c.relowner) AS sequenceowner, "
                 "s.data_type::regtype AS data_type, "
                 "s.start_value AS start_value, "
                 "s.minimum_value AS min_value, "
                 "s.maximum_value AS max_value, "
                 "s.increment AS increment_by, "
                 "s.cycle_option AS cycle "
                 "FROM information_schema.sequences s "
                 "JOIN pg_class c ON c.relname = s.sequence_name "
                 "LEFT JOIN pg_namespace n ON n.oid = c.relnamespace "
                 "WHERE NOT pg_is_other_temp_schema(n.oid) "
                 "AND c.relkind = 'S'::\"char\" "
                 "AND sequence_name = %(name)s "
                 "AND sequence_schema = %(schema)s")

        res = exec_sql(self, query,
                       query_params={'name': self.name, 'schema': self.schema},
                       add_to_executed=False)

        if not res:
            self.exists = False
            return False

        if res:
            self.exists = True
            self.schema = res[0]['schemaname']
            self.name = res[0]['sequencename']
            self.owner = res[0]['sequenceowner']
            self.data_type = res[0]['data_type']
            self.start_value = res[0]['start_value']
            self.minvalue = res[0]['min_value']
            self.maxvalue = res[0]['max_value']
            self.increment = res[0]['increment_by']
            self.cycle = res[0]['cycle']

    def create(self):
        """Implements CREATE SEQUENCE command behavior."""
        query = ['CREATE SEQUENCE']
        query.append(self.__add_schema())

        if self.module.params.get('data_type'):
            query.append('AS %s' % self.module.params['data_type'])

        if self.module.params.get('increment'):
            query.append('INCREMENT BY %s' % self.module.params['increment'])

        if self.module.params.get('minvalue'):
            query.append('MINVALUE %s' % self.module.params['minvalue'])

        if self.module.params.get('maxvalue'):
            query.append('MAXVALUE %s' % self.module.params['maxvalue'])

        if self.module.params.get('start'):
            query.append('START WITH %s' % self.module.params['start'])

        if self.module.params.get('cache'):
            query.append('CACHE %s' % self.module.params['cache'])

        if self.module.params.get('cycle'):
            query.append('CYCLE')

        return exec_sql(self, ' '.join(query), return_bool=True)

    def drop(self):
        """Implements DROP SEQUENCE command behavior."""
        query = ['DROP SEQUENCE']
        query.append(self.__add_schema())

        if self.module.params.get('cascade'):
            query.append('CASCADE')

        return exec_sql(self, ' '.join(query), return_bool=True)

    def rename(self):
        """Implements ALTER SEQUENCE RENAME TO command behavior."""
        query = ['ALTER SEQUENCE']
        query.append(self.__add_schema())
        query.append('RENAME TO "%s"' % self.module.params['rename_to'])

        return exec_sql(self, ' '.join(query), return_bool=True)

    def set_owner(self):
        """Implements ALTER SEQUENCE OWNER TO command behavior."""
        query = ['ALTER SEQUENCE']
        query.append(self.__add_schema())
        query.append('OWNER TO "%s"' % self.module.params['owner'])

        return exec_sql(self, ' '.join(query), return_bool=True)

    def set_schema(self):
        """Implements ALTER SEQUENCE SET SCHEMA command behavior."""
        query = ['ALTER SEQUENCE']
        query.append(self.__add_schema())
        query.append('SET SCHEMA "%s"' % self.module.params['newschema'])

        return exec_sql(self, ' '.join(query), return_bool=True)

    def __add_schema(self):
        return '"%s"."%s"' % (self.schema, self.name)


# ===========================================
# Module execution.
#

def main():
    argument_spec = postgres_common_argument_spec()
    argument_spec.update(
        sequence=dict(type='str', required=True, aliases=['name']),
        state=dict(type='str', default='present', choices=['absent', 'present']),
        data_type=dict(type='str', choices=['bigint', 'integer', 'smallint']),
        increment=dict(type='int'),
        minvalue=dict(type='int', aliases=['min']),
        maxvalue=dict(type='int', aliases=['max']),
        start=dict(type='int'),
        cache=dict(type='int'),
        cycle=dict(type='bool', default=False),
        schema=dict(type='str', default='public'),
        cascade=dict(type='bool', default=False),
        rename_to=dict(type='str'),
        owner=dict(type='str'),
        newschema=dict(type='str'),
        db=dict(type='str', default='', aliases=['login_db', 'database']),
        session_role=dict(type='str'),
        trust_input=dict(type="bool", default=True),
    )
    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        mutually_exclusive=[
            ['rename_to', 'data_type'],
            ['rename_to', 'increment'],
            ['rename_to', 'minvalue'],
            ['rename_to', 'maxvalue'],
            ['rename_to', 'start'],
            ['rename_to', 'cache'],
            ['rename_to', 'cycle'],
            ['rename_to', 'cascade'],
            ['rename_to', 'owner'],
            ['rename_to', 'newschema'],
            ['cascade', 'data_type'],
            ['cascade', 'increment'],
            ['cascade', 'minvalue'],
            ['cascade', 'maxvalue'],
            ['cascade', 'start'],
            ['cascade', 'cache'],
            ['cascade', 'cycle'],
            ['cascade', 'owner'],
            ['cascade', 'newschema'],
        ]
    )

    if not module.params["trust_input"]:
        check_input(
            module,
            module.params['sequence'],
            module.params['schema'],
            module.params['rename_to'],
            module.params['owner'],
            module.params['newschema'],
            module.params['session_role'],
        )

    # Note: we don't need to check mutually exclusive params here, because they are
    # checked automatically by AnsibleModule (mutually_exclusive=[] list above).

    # Change autocommit to False if check_mode:
    autocommit = not module.check_mode
    # Connect to DB and make cursor object:
    conn_params = get_conn_params(module, module.params)
    db_connection = connect_to_db(module, conn_params, autocommit=autocommit)
    cursor = db_connection.cursor(cursor_factory=DictCursor)

    ##############
    # Create the object and do main job:
    data = Sequence(module, cursor)

    # Set defaults:
    changed = False

    # Create new sequence
    if not data.exists and module.params['state'] == 'present':
        if module.params.get('rename_to'):
            module.fail_json(msg="Sequence '%s' does not exist, nothing to rename" % module.params['sequence'])
        if module.params.get('newschema'):
            module.fail_json(msg="Sequence '%s' does not exist, change of schema not possible" % module.params['sequence'])

        changed = data.create()

    # Drop non-existing sequence
    elif not data.exists and module.params['state'] == 'absent':
        # Nothing to do
        changed = False

    # Drop existing sequence
    elif data.exists and module.params['state'] == 'absent':
        changed = data.drop()

    # Rename sequence
    if data.exists and module.params.get('rename_to'):
        if data.name != module.params['rename_to']:
            changed = data.rename()
            if changed:
                data.new_name = module.params['rename_to']

    # Refresh information
    if module.params['state'] == 'present':
        data.get_info()

    # Change owner, schema and settings
    if module.params['state'] == 'present' and data.exists:
        # change owner
        if module.params.get('owner'):
            if data.owner != module.params['owner']:
                changed = data.set_owner()

        # Set schema
        if module.params.get('newschema'):
            if data.schema != module.params['newschema']:
                changed = data.set_schema()
                if changed:
                    data.new_schema = module.params['newschema']

    # Rollback if it's possible and check_mode:
    if module.check_mode:
        db_connection.rollback()
    else:
        db_connection.commit()

    cursor.close()
    db_connection.close()

    # Make return values:
    kw = dict(
        changed=changed,
        state='present',
        sequence=data.name,
        queries=data.executed_queries,
        schema=data.schema,
        data_type=data.data_type,
        increment=data.increment,
        minvalue=data.minvalue,
        maxvalue=data.maxvalue,
        start=data.start_value,
        cycle=data.cycle,
        owner=data.owner,
    )

    if module.params['state'] == 'present':
        if data.new_name:
            kw['newname'] = data.new_name
        if data.new_schema:
            kw['newschema'] = data.new_schema

    elif module.params['state'] == 'absent':
        kw['state'] = 'absent'

    module.exit_json(**kw)


if __name__ == '__main__':
    main()
